import warnings
from functools import wraps
from typing import Callable, List, Union

import numpy as np
from .geometry import Geometry

__all__ = ["GeometryProperties"]


def get_mesh(func):
    @wraps(func)
    def wrapper(geometry: Geometry, *args, **kwargs):
        fit = geometry.model
        if not hasattr(fit, "mesh"):
            return None
        return func(fit, *args, **kwargs)

    return wrapper


@get_mesh
def mesh_curvature(fit, curvature: str, radius: int, **kwargs):
    return fit.compute_curvature(curvature=curvature, radius=radius, **kwargs)


@get_mesh
def mesh_edge_length(fit, **kwargs):
    from .meshing.utils import compute_edge_lengths

    return compute_edge_lengths(fit.mesh)


@get_mesh
def mesh_surface_area(fit, **kwargs):
    return fit.mesh.get_surface_area()


@get_mesh
def mesh_volume(fit, **kwargs):
    return fit.mesh.get_volume()


@get_mesh
def mesh_triangle_area(fit, **kwargs):
    vertices, triangles = fit.vertices, fit.triangles
    v0 = vertices[triangles[:, 0]]
    v1 = vertices[triangles[:, 1]]
    v2 = vertices[triangles[:, 2]]
    return np.linalg.norm(np.cross(v1 - v0, v2 - v0), axis=1) / 2


@get_mesh
def mesh_triangle_volume(fit, **kwargs):
    vertices, triangles = fit.vertices, fit.triangles
    v0 = vertices[triangles[:, 0]]
    v1 = vertices[triangles[:, 1]]
    v2 = vertices[triangles[:, 2]]
    face_volumes = np.sum(np.cross(v0, v1) * v2, axis=1) / 6.0
    return np.array([np.sum(np.abs(face_volumes))])


@get_mesh
def mesh_vertices(fit, **kwargs):
    return fit.vertices.shape[0]


@get_mesh
def mesh_triangles(fit, **kwargs):
    return fit.triangles.shape[0]


def distance(
    geometry: Geometry,
    queries: List[Union[np.ndarray, Geometry]] = [],
    k: int = 1,
    k_start: int = 1,
    include_self: bool = False,
    only_self: bool = False,
    *args,
    **kwargs,
):
    from mosaic.utils import find_closest_points

    if not isinstance(queries, (list, tuple)):
        queries = [queries]

    if only_self:
        queries, include_self = [geometry], True

    distance = None
    for query in queries:
        if not include_self and id(query) == id(geometry):
            continue

        if isinstance(query, Geometry):
            dist = query.compute_distance(geometry.points, k=k)
        else:
            dist, _ = find_closest_points(query, geometry.points, k=k)

        if distance is None:
            distance = dist
        distance = np.minimum(distance, dist)

    if distance is None:
        return None

    k_start = max(k_start - 1, 0)
    if k_start == k:
        raise ValueError("k_start needs to be smaller than k")

    if distance.ndim == 2:
        distance = distance[(slice(None), slice(k_start, k))].mean(axis=1)

    return distance


def box_size(geometry, axis: int = None):
    shape = geometry.points.max(axis=0) - geometry.points.min(axis=0)
    if axis is None:
        return shape
    return shape[axis]


def width(geometry, *args, **kwargs):
    return geometry.points[:, 0]


def depth(geometry, *args, **kwargs):
    return geometry.points[:, 1]


def height(geometry, *args, **kwargs):
    return geometry.points[:, 2]


def n_points(geometry, *args, **kwargs):
    return geometry.points.shape[0]


def projected_curvature(
    geometry: Geometry, queries: List[Geometry], curvature: str, radius: int, **kwargs
):
    if len(queries) == 0:
        return None
    elif len(queries) > 1:
        warnings.warn("Using the first query instance.")

    if (fit := queries[0].model) is None:
        return None

    curvature = fit.compute_curvature(curvature=curvature, radius=radius, **kwargs)
    _, indices = fit.compute_distance(points=geometry.points, return_indices=True)
    return curvature[indices]


def geodesic_distance(
    geometry: Geometry, queries: List[Geometry], k: int = 1, k_start=1
):
    if len(queries) == 0:
        return None
    elif len(queries) > 1:
        warnings.warn("Using the first query instance.")

    if (fit := queries[0].model) is None:
        return None

    _, indices = fit.compute_distance(points=geometry.points, return_indices=True)
    distance = fit.geodesic_distance(
        target_vertices=indices, source_vertices=indices, k=k
    )

    k_start = max(k_start - 1, 0)
    if distance.ndim == 2:
        distance = distance[(slice(None), slice(k_start, k))].mean(axis=1)

    return distance


def vertex_property(geometry, name: str, *args, **kwargs):
    if geometry.vertex_properties is None:
        return None
    # We copy as safeguard as values may be modified
    return geometry.vertex_properties.get_property(name).copy()


class GeometryProperties:
    """Registry for property calculators."""

    _calculators = {
        "distance": distance,
        "box_size": box_size,
        "width": width,
        "depth": depth,
        "height": height,
        "n_points": n_points,
        "mesh_curvature": mesh_curvature,
        "mesh_edge_length": mesh_edge_length,
        "mesh_surface_area": mesh_surface_area,
        "mesh_triangle_area": mesh_triangle_area,
        "mesh_volume": mesh_volume,
        "mesh_triangle_volume": mesh_triangle_volume,
        "mesh_vertices": mesh_vertices,
        "mesh_triangles": mesh_triangles,
        "projected_curvature": projected_curvature,
        "geodesic_distance": geodesic_distance,
        "vertex_property": vertex_property,
    }

    @classmethod
    def register(cls, property_name: str, func: Callable):
        """Register a calculator function for a specific property."""
        cls._calculators[property_name] = func

    @classmethod
    def compute(cls, property_name, *args, **kwargs):
        """Compute a property for a geometry object."""
        func = cls._calculators.get(property_name, None)
        if func is None:
            print(f"Unknown property: {property_name}")
            return None

        return func(*args, **kwargs)



from uuid import uuid4
from typing import Dict, List
from dataclasses import dataclass, field

from qtpy.QtGui import QColor, QIcon, QPixmap, QPainter
from qtpy.QtCore import Qt, QRect, QByteArray, QItemSelection, QItemSelectionModel
from qtpy.QtWidgets import (
    QFrame,
    QVBoxLayout,
    QTreeWidget,
    QTreeWidgetItem,
    QApplication,
    QStyledItemDelegate,
    QStyle,
    QAbstractItemView,
)
from qtpy.QtSvg import QSvgRenderer
import qtawesome as qta


@dataclass()
class TreeState:
    """Minimal tree structure tracking."""

    #: {'Group 1': ['uuid1', 'uuid2'], ...}
    groups: Dict[str, List[str]] = field(default_factory=dict)
    #: {'Group 1', 'uuid1', ...}
    root_order: Dict[str, int] = field(default_factory=dict)
    #: ['uuid3', 'uuid4', ...]
    root_items: List[str] = field(default_factory=list)

    def get_all_uuids(self):
        """Get all UUIDs currently in the tree."""
        uuids = set(self.root_items)
        for group_uuids in self.groups.values():
            uuids.update(group_uuids)
        return uuids


class ContainerTreeWidget(QFrame):
    """Drop-in replacement for ContainerListWidget using QTreeWidget for grouping support."""

    def __init__(self, title: str = None, border: bool = True):
        super().__init__()
        self.setFrameStyle(QFrame.Shape.NoFrame)

        layout = QVBoxLayout(self)
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)

        self.title = title
        app = QApplication.instance()
        app.paletteChanged.connect(self.update_style)

        self.tree_widget = QTreeWidget()
        self.tree_widget.setHeaderHidden(True)
        self.tree_widget.setHorizontalScrollBarPolicy(
            Qt.ScrollBarPolicy.ScrollBarAlwaysOff
        )
        self.tree_widget.setIndentation(5)
        self.tree_widget.setAnimated(True)
        self.tree_widget.setRootIsDecorated(False)
        self.tree_widget.setItemsExpandable(True)
        self.tree_widget.setExpandsOnDoubleClick(False)

        self.tree_widget.itemClicked.connect(self._on_item_clicked)

        self.tree_widget.setDragEnabled(False)
        self.tree_widget.setAcceptDrops(True)
        self.tree_widget.setDropIndicatorShown(True)
        self.tree_widget.setDragDropMode(QAbstractItemView.DragDropMode.InternalMove)

        self.tree_widget.setItemDelegate(MetadataItemDelegate(self.tree_widget))

        self.tree_widget.setStyleSheet(
            """
            QTreeWidget {
                border: none;
                background-color: transparent;
                outline: none;
                padding: 4px 0px;
                font-size: 13px;
            }
            QTreeWidget::item {
                border-radius: 6px;
                border: none;
                padding: 4px 0px;
                margin: 2px 0px;
                outline: none;
            }
            QTreeWidget::item:hover {
                background-color: rgba(0, 0, 0, 0.0);
            }
            QTreeWidget::item:selected {
                background-color: rgba(0, 0, 0, 0.0);
                font-weight: 500;
            }
            QLineEdit {
                background-color: white;
                border: 1px solid #4f46e5;
                border-radius: 6px;
                padding: 0px 3px;
                margin: 0px 8px;
                selection-background-color: rgba(99, 102, 241, 0.6);
                font-size: 13px;
            }
        """
        )

        layout.addWidget(self.tree_widget)
        if border:
            self.update_style()

    def selected_items(self):
        # We specifically omit GroupTreeWidgetItem
        return [
            item
            for item in self.tree_widget.selectedItems()
            if isinstance(item, StyledTreeWidgetItem)
        ]

    def update_style(self):
        return self.setStyleSheet(
            """
            QFrame {
                background-color: transparent;
                border: none;
                border-bottom: 1px solid #6b7280;
            }
        """
        )

    def to_state(self) -> TreeState:
        """Extract current tree structure as TreeState object."""
        state = TreeState()

        for item, parent, _ in self.traverse(reverse=False):

            # GroupTreeWidgetItem structure is evident from parent
            if not isinstance(item, StyledTreeWidgetItem):
                continue

            if (uuid := item.metadata.get("uuid")) is None:
                continue

            group_name = getattr(parent, "group_name", None)

            if parent is None:
                state.root_items.append(uuid)
            elif isinstance(parent, GroupTreeWidgetItem):
                if group_name not in state.groups:
                    state.groups[group_name] = []
                state.groups[group_name].append(uuid)
                uuid = parent.metadata.get("uuid")

            if uuid not in state.root_order:
                state.root_order[uuid] = (len(state.root_order), group_name)

        return state

    def apply_state(self, state: TreeState, uuid_to_items: Dict):
        """Apply tree structure to existing items.

        Parameters
        ----------
        state : :py:class:`TreeState`
            Desired tree structure
        uuid_to_items : dict
            Map of UUID to QTreeWidgetItem
        """
        self.tree_widget.clear()

        order = dict(sorted(state.root_order.items(), key=lambda x: x[1][0]))

        for uuid, (_, group_name) in order.items():
            if group_name is None:
                self.tree_widget.addTopLevelItem(uuid_to_items[uuid])
                continue

            group_item = self.create_group(group_name)
            uuids = [x for x in state.groups.get(group_name, []) if x in uuid_to_items]
            for uuid in uuids:
                group_item.addChild(uuid_to_items[uuid])

    def update(self, uuid_to_items):
        """
        Update tree incrementally based on provided items.

        Parameters
        ----------
        uuid_to_items : dict
            Map from UUID to QTreeWidgetItem to be added/updated
        """
        try:
            self.tree_widget.blockSignals(True)
            existing_uuids = self._process_tree_items(uuid_to_items)
        finally:
            self.tree_widget.blockSignals(False)
        for uuid, item in uuid_to_items.items():
            if uuid in existing_uuids:
                continue
            self.tree_widget.addTopLevelItem(item)

    def _move_items_to_parent(self, items, new_parent):
        """Move items to a new parent (or root if None).

        Parameters
        ----------
        items : list of QTreeWidgetItem
            Items to move
        new_parent : GroupTreeWidgetItem or None
            New parent, or None for root level
        """
        for item in items:
            if old_parent := item.parent():
                old_parent.removeChild(item)
            else:
                index = self.tree_widget.indexOfTopLevelItem(item)
                self.tree_widget.takeTopLevelItem(index)

            if new_parent:
                new_parent.addChild(item)
            else:
                self.tree_widget.addTopLevelItem(item)

    def group_selected(self, group_name: str):
        """Create a new group with currently selected items.

        Parameters
        ----------
        group_name : str
            Name for the new group

        Returns
        -------
        GroupTreeWidgetItem or None
            The created group item, or None if no items selected
        """
        if not (selected_items := self.selected_items()):
            return None

        first_item = selected_items[0]
        insert_index = self.tree_widget.indexOfTopLevelItem(first_item)

        group_item = self.create_group(group_name, insert_index=insert_index)
        try:
            self.tree_widget.blockSignals(True)
            self._move_items_to_parent(selected_items, group_item)
        finally:
            self.tree_widget.blockSignals(False)

        group_item.setExpanded(True)
        self._select_group_children(group_item)
        return group_item

    def ungroup_selected(self) -> int:
        """Move selected items to root level (removing them from their groups).

        Returns
        -------
        int
            Number of items ungrouped
        """
        if not (selected_items := self.selected_items()):
            return 0

        try:
            self.tree_widget.blockSignals(True)
            self._move_items_to_parent(selected_items, None)
        finally:
            self.tree_widget.blockSignals(False)

        self._set_selection(selected_items)
        return len(selected_items)

    def traverse(self, reverse=False):
        """Generator that yields all (item, parent, index) tuples.

        Parameters
        ----------
        reverse : bool
            If True, iterate in reverse order (useful for mutations)
        """
        items = []

        # Collect all items with their metadata
        for i in range(self.tree_widget.topLevelItemCount()):
            item = self.tree_widget.topLevelItem(i)
            items.append((item, None, i))

            if isinstance(item, GroupTreeWidgetItem):
                for j in range(item.childCount()):
                    child = item.child(j)
                    items.append((child, item, j))

        # Yield in requested order
        if reverse:
            yield from reversed(items)
        else:
            yield from items

    def _process_tree_items(self, uuid_to_items):
        """Walk tree, replace existing items, and remove invalid items."""
        existing_uuids = set()

        for item, parent, index in self.traverse(reverse=True):
            if isinstance(item, StyledTreeWidgetItem):
                uuid = item.metadata.get("uuid")
                # Remove non existing items
                if uuid not in uuid_to_items:
                    if parent is not None:
                        parent.removeChild(item)
                    else:
                        self.tree_widget.takeTopLevelItem(index)
                    continue

                # Update visibility status and metadata
                item.update(uuid_to_items[uuid])
                existing_uuids.add(uuid)

            # Remove empty groups
            elif isinstance(item, GroupTreeWidgetItem):
                if item.childCount() == 0:
                    self.tree_widget.takeTopLevelItem(index)
        return existing_uuids

    def __getattr__(self, name):
        """Forward all other attributes to tree_widget for compatibility."""
        return getattr(self.tree_widget, name)

    def addItem(self, item):
        self.tree_widget.addTopLevelItem(item)

    def create_group(self, name: str, insert_index: int = None):
        """Create a new group at the root level.

        Parameters
        ----------
        name : str
            Name for the new group
        insert_index : int, optional
            Index at which to insert the group. If None, appends to end.
        """
        group_item = GroupTreeWidgetItem(name)
        if insert_index is not None and insert_index >= 0:
            self.tree_widget.insertTopLevelItem(insert_index, group_item)
        else:
            self.tree_widget.addTopLevelItem(group_item)
        group_item.setExpanded(True)
        return group_item

    def _on_item_clicked(self, item, column):
        """Handle item clicks - toggle expand/collapse for groups and select children."""
        if not isinstance(item, GroupTreeWidgetItem):
            return

        cursor_pos = self.tree_widget.mapFromGlobal(self.tree_widget.cursor().pos())
        item_rect = self.tree_widget.visualItemRect(item)

        # If clicking on arrow area, toggle expand/collapse
        if (cursor_pos.x() - item_rect.left()) <= 40:
            item.setExpanded(not item.isExpanded())
            item.update_icon(item.isExpanded())
        self._select_group_children(item)

    def _set_selection(self, items):
        """Set selection to specific items.

        Parameters
        ----------
        items : list of QTreeWidgetItem or single QTreeWidgetItem
            Items to select
        """
        if not isinstance(items, (list, tuple)):
            items = [items]

        selection = QItemSelection()
        for item in items:
            if item is None:
                continue
            index = self.tree_widget.indexFromItem(item)
            selection.select(index, index)

        self.tree_widget.selectionModel().select(
            selection, QItemSelectionModel.SelectionFlag.ClearAndSelect
        )

    def _select_group_children(self, group_item):
        """Select all children of a group and the group itself."""
        if not isinstance(group_item, GroupTreeWidgetItem):
            return None

        items_to_select = [group_item]

        for i in range(group_item.childCount()):
            child = group_item.child(i)
            if isinstance(child, StyledTreeWidgetItem):
                items_to_select.append(child)

        self._set_selection(items_to_select)


class GroupTreeWidgetItem(QTreeWidgetItem):
    """Special tree widget item representing a group."""

    def __init__(self, name: str, parent=None):
        super().__init__(parent, [name])
        self.group_name = name
        self.arrow_color = "#6b7280"

        self.update_icon()

        # Groups can be renamed but not dragged
        self.setFlags(
            Qt.ItemFlag.ItemIsEnabled
            | Qt.ItemFlag.ItemIsDropEnabled
            | Qt.ItemFlag.ItemIsEditable
            | Qt.ItemFlag.ItemIsSelectable
        )
        self.metadata = {"uuid": str(uuid4())}

    def update_icon(self, expanded: bool = True):
        """Update the icon based on expanded state."""

        path = "M7,5 L11,9 L7,13"
        if expanded:
            path = "M5,7 L9,11 L13,7"

        svg_template = f"""
            <svg xmlns="http://www.w3.org/2000/svg" viewBox="0 0 18 18">
                <rect width="18" height="18" fill="transparent" />
                <path stroke="{self.arrow_color}" stroke-width="2" fill="none" d="{path}" />
            </svg>"""

        svg_bytes = QByteArray(svg_template.encode())
        renderer = QSvgRenderer(svg_bytes)
        pixmap = QPixmap(18, 18)
        pixmap.fill(Qt.GlobalColor.transparent)
        painter = QPainter(pixmap)
        renderer.render(painter)
        painter.end()
        icon = QIcon(pixmap)
        self.setIcon(0, icon)

    def setData(self, column, role, value):
        """Update group_name when text is changed."""
        if role == Qt.ItemDataRole.EditRole:
            self.group_name = value
        return super().setData(column, role, value)


class StyledTreeWidgetItem(QTreeWidgetItem):
    """
    Create a styled tree widget item with type-specific icons.

    Parameters
    ----------
    text : str
        The display text for the item
    visible : bool
        Whether the item is visible
    metadata : dict
        Additional metadata for the item
    parent : QWidget or QTreeWidgetItem
        Parent widget or parent tree item
    editable : bool
        Whether the item is editable
    """

    def __init__(self, text, visible=True, metadata=None, parent=None, editable=False):

        super().__init__(parent, [text])

        self.original_color = self.foreground(0)
        self.visible_color = QColor(99, 102, 241)
        self.invisible_color = QColor(128, 128, 128)

        self.metadata = metadata or {}

        _ = self.metadata.pop("metadata_text", None)
        if editable:
            self.setFlags(self.flags() | Qt.ItemFlag.ItemIsEditable)

        # Items can be dragged and selected, but do not accept drops
        # to prevent creating hierarchies of StyledTreeWidgetItem
        self.setFlags(
            self.flags() | Qt.ItemFlag.ItemIsDragEnabled | Qt.ItemFlag.ItemIsSelectable
        )
        self.setFlags(self.flags() & ~Qt.ItemFlag.ItemIsDropEnabled)

        self.set_visible(visible)

    def update(self, other: "StyledTreeWidgetItem"):
        if other is None:
            return None

        self.metadata = other.metadata.copy()
        self.setText(0, other.text())

        self.set_visible(other.visible)

    def update_icon(self, visible):
        """Update the item icon based on type and visibility."""
        self.visible = visible

        item_type = self.metadata.get("item_type")
        if item_type == "cluster":
            icon_name = "mdi.scatter-plot"
        elif item_type == "parametric":
            icon_name = "mdi.function"
        elif item_type == "mesh":
            icon_name = "mdi.triangle-outline"
        elif item_type == "trajectory":
            icon_name = "mdi.chart-line-variant"
        else:
            icon_name = "mdi.shape-outline"

        color = self.visible_color if visible else self.invisible_color
        icon = qta.icon(icon_name, color=color, scale_factor=0.85)
        self.setIcon(0, icon)

    def set_visible(self, visible):
        """Update visibility state and icon."""
        self.update_icon(visible)
        self.setForeground(0, self.original_color if visible else self.invisible_color)

    def text(self):
        """Get item text for backward compatibility."""
        return super().text(0)

    def setData(self, *args):
        if len(args) == 2:
            index, (column, value) = 0, args
        elif len(args) == 3:
            index, column, value = args
        else:
            return None
        return super().setData(index, column, value)

    def data(self, *args):
        if len(args) == 1:
            index, column = 0, *args
        elif len(args) == 2:
            index, column = args
        else:
            return None
        return super().data(index, column)


class MetadataItemDelegate(QStyledItemDelegate):
    """Delegate for custom selection/hover painting."""

    def __init__(self, parent=None):
        super().__init__(parent)

    def paint(self, painter, option, index):
        tree_widget = self.parent()
        item = tree_widget.itemFromIndex(index)

        # Calculate content rect extending to right edge
        content_rect = QRect(
            option.rect.left() + 6,
            option.rect.top() + 2,
            option.rect.width() - 6,
            option.rect.height() - 4,
        )

        # Draw hover/selection background
        painter.save()
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)
        if option.state & QStyle.StateFlag.State_Selected:
            painter.setBrush(QColor(99, 102, 241, int(0.3 * 255)))
            painter.setPen(Qt.PenStyle.NoPen)
            painter.drawRoundedRect(content_rect, 6, 6)
        elif option.state & QStyle.StateFlag.State_MouseOver:
            painter.setBrush(QColor(0, 0, 0, int(0.1 * 255)))
            painter.setPen(Qt.PenStyle.NoPen)
            painter.drawRoundedRect(content_rect, 6, 6)
        painter.restore()

        # Draw icon
        icon_size = 20
        icon = index.data(Qt.ItemDataRole.DecorationRole)
        if icon and not icon.isNull():
            icon_rect = QRect(
                option.rect.left() + 12,
                option.rect.top() + (option.rect.height() - icon_size) // 2,
                icon_size,
                icon_size,
            )
            icon.paint(painter, icon_rect)

        # Draw text
        painter.save()
        text = index.data(Qt.ItemDataRole.DisplayRole)
        if isinstance(item, StyledTreeWidgetItem) and not item.visible:
            painter.setPen(QColor(128, 128, 128))
        else:
            painter.setPen(option.palette.color(option.palette.ColorRole.Text))

        text_rect = QRect(
            option.rect.left() + 12 + icon_size + 4,
            option.rect.top(),
            option.rect.width() - icon_size - 28,
            option.rect.height(),
        )
        painter.drawText(text_rect, int(Qt.AlignmentFlag.AlignVCenter), text)
        painter.restore()


# Backward compatibility aliases
ContainerListWidget = ContainerTreeWidget
StyledListWidgetItem = StyledTreeWidgetItem



"""
Implements geometric surface models for point cloud data. This includes
parameteric as well as non-parametric triangular-mesh based approaches.

Children of the underlying abstract Parametrization class, also define
means for equidistant sampling and computation of normal vectors.
Furthermore, there are amenable to native python pickling.

Copyright (c) 2023-2025 European Molecular Biology Laboratory

Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

import warnings
from typing import Tuple
from abc import ABC, abstractmethod

import igl
import numpy as np
import open3d as o3d
from scipy import optimize, interpolate

from .utils import (
    find_closest_points,
    com_cluster_points,
    compute_normals,
    points_to_volume,
)
from .meshing import (
    triangulate_refine_fair,
    fair_mesh,
    remesh,
    to_open3d,
    poisson_mesh,
)

__all__ = [
    "Sphere",
    "Ellipsoid",
    "Cylinder",
    "RBF",
    "TriangularMesh",
    "PoissonMesh",
    "ClusteredBallPivotingMesh",
    "ConvexHull",
    "FlyingEdges",
    "SplineCurve",
]


def _sample_from_mesh(mesh, n_samples: int, mesh_init_factor: int = None) -> np.ndarray:
    if mesh_init_factor is None:
        point_cloud = mesh.sample_points_uniformly(
            number_of_points=n_samples,
        )
    else:
        point_cloud = mesh.sample_points_poisson_disk(
            number_of_points=n_samples,
            init_factor=mesh_init_factor,
        )
    return np.asarray(point_cloud.points)


def _sample_from_chull(
    positions_xyz: np.ndarray, n_samples: int, mesh_init_factor: int = None
) -> np.ndarray:
    chull = ConvexHull.fit(positions_xyz)
    return _sample_from_mesh(chull.mesh, n_samples, mesh_init_factor)


def _normalize(arr: np.ndarray):
    arr = np.atleast_2d(arr)
    norm = np.linalg.norm(arr, axis=1, keepdims=True)
    norm = np.where(norm > 1e-6, norm, 1)
    return np.divide(arr, norm, out=arr)


class Parametrization(ABC):
    """Abstract base class to represent picklable parametrizations."""

    def __init__(self, *args, **kwargs):
        pass

    @abstractmethod
    def fit(self, positions: np.ndarray, *args, **kwargs) -> "Parametrization":
        """
        Fit a parametrization to a point cloud.

        Parameters
        ----------
        positions : np.ndarray
            Point coordinates with shape (n x 3)
        *args : List
            Additional arguments
        **kwargs : Dict
            Additional keywoard arguments.

        Returns
        -------
        Parametrization
            Parametrization instance.
        """

    @abstractmethod
    def sample(self, n_samples: int, normal_offset: float = 0.0, *args, **kwargs):
        """
        Samples points from the surface of the parametrization.

        Parameters
        ----------
        n_samples : int
            Number of samples to draw
        normal_offset : float, optional
            Offset points by normal_offset times their normal vector.
        *args : List
            Additional arguments
        **kwargs : Dict
            Additional keywoard arguments.

        Returns
        -------
        np.ndarray
            Point coordinates (n, 3).
        """

    @abstractmethod
    def compute_normal(self, positions: np.ndarray, *args, **kwargs):
        """
        Compute the normal vector at a given point on the surface.

        Parameters
        ----------
        points : np.ndarray
            Points on the surface with shape n x d

        Returns
        -------
        np.ndarray
            Normal vectors at the given points
        """

    @abstractmethod
    def points_per_sampling(
        self, sampling_density: float, normal_offset: float = None
    ) -> int:
        """
        Computes the approximate number of random samples
        required to achieve a given spatial sampling_density.

        Parameters
        ----------
        sampling_density : float
            Average distance between points.
        normal_offset : float, optional
            Compute number of samples on offset parametrization, instead of current.

        Returns
        -------
        int
            Number of required random samples.
        """

    def compute_distance(self, points: np.ndarray, **kwargs) -> np.ndarray:
        """
        Computes the distance between points and the parameterization.

        Parameters
        ----------
        points : np.ndarray
            Array of coordinates (n, d).

        Returns
        -------
        np.ndarray
            Distances between points and the parametrization.
        """
        samples = self.sample(n_samples=points.shape[0] * 4)
        distances, _ = find_closest_points(samples, points, k=1)
        return distances


class Sphere(Parametrization):
    """
    Parametrize a point cloud as sphere.

    Parameters
    ----------
    radius : np.ndarray
        Radius of the sphere
    center : np.ndarray
        Center of the sphere along each axis.
    """

    def __init__(self, radius: np.ndarray, center: np.ndarray):
        self.radius = radius
        self.center = center

    @classmethod
    def fit(cls, positions: np.ndarray, **kwargs) -> "Sphere":
        positions = np.asarray(positions, dtype=np.float64)
        A = np.column_stack((2 * positions, np.ones(len(positions))))
        b = (positions**2).sum(axis=1)

        x, res, _, _ = np.linalg.lstsq(A, b, rcond=None)

        radius = np.sqrt(x[0] ** 2 + x[1] ** 2 + x[2] ** 2 + x[3])
        return cls(radius=radius, center=x[:3])

    def sample(
        self, n_samples: int, normal_offset: float = 0.0, **kwargs
    ) -> np.ndarray:
        """
        Samples points from the surface of a sphere.

        Parameters
        ----------
        n_samples : int
            Number of samples to draw
        normal_offset : float, optional
            Offset points by normal_offset times their normal vector.

        Returns
        -------
        np.ndarray
            Point coordinates (n, 3).
        """
        radius = self.radius

        indices = np.arange(0, n_samples, dtype=float) + 0.5
        phi = np.arccos(1 - 2 * indices / n_samples)
        theta = np.pi * (1 + 5**0.5) * indices

        if normal_offset is not None:
            radius = radius + normal_offset

        positions_xyz = np.column_stack(
            [np.cos(theta) * np.sin(phi), np.sin(theta) * np.sin(phi), np.cos(phi)]
        )
        positions_xyz = np.multiply(positions_xyz, radius)
        return np.add(positions_xyz, self.center)

    def compute_normal(self, points: np.ndarray) -> np.ndarray:
        normals = (points - self.center) / self.radius
        return _normalize(normals)

    def compute_distance(self, points: np.ndarray, **kwargs) -> np.ndarray:
        centered = np.linalg.norm(points - self.center, axis=1)
        return np.abs(centered - self.radius)

    def points_per_sampling(
        self, sampling_density: float, normal_offset: float = None
    ) -> int:
        radius = self.radius
        if normal_offset is not None:
            radius = radius + normal_offset
        return int(4 * np.ceil(np.power(radius / sampling_density, 2)))


class Ellipsoid(Parametrization):
    """
    Parametrize a point cloud as ellipsoid.

    Parameters
    ----------
    radii : np.ndarray
        Radii of the ellipse along each axis
    center : np.ndarray
        Center of the ellipse along each axis
    orientations : np.ndarray
        Square orientation matrix
    """

    def __init__(self, radii: np.ndarray, center: np.ndarray, orientations: np.ndarray):
        self.radii = np.asarray(radii)
        self.center = np.asarray(center)
        self.orientations = np.asarray(orientations)

    @classmethod
    def fit(cls, positions, **kwargs) -> "Ellipsoid":
        # Adapted from https://de.mathworks.com/matlabcentral/fileexchange/24693-ellipsoid-fit
        positions = np.asarray(positions, dtype=np.float64)
        if positions.shape[1] != 3 or len(positions.shape) != 2:
            raise NotImplementedError(
                "Only three-dimensional point clouds are supported."
            )

        x, y, z = positions[:, 0], positions[:, 1], positions[:, 2]
        D = np.array(
            [
                x * x + y * y - 2 * z * z,
                x * x + z * z - 2 * y * y,
                2 * x * y,
                2 * x * z,
                2 * y * z,
                2 * x,
                2 * y,
                2 * z,
                1 - 0 * x,
            ]
        )
        d2 = np.array(x * x + y * y + z * z).T
        u = np.linalg.solve(D.dot(D.T), D.dot(d2))
        v = np.concatenate(
            [
                np.array([u[0] + 1 * u[1] - 1]),
                np.array([u[0] - 2 * u[1] - 1]),
                np.array([u[1] - 2 * u[0] - 1]),
                u[2:],
            ],
            axis=0,
        ).flatten()
        A = np.array(
            [
                [v[0], v[3], v[4], v[6]],
                [v[3], v[1], v[5], v[7]],
                [v[4], v[5], v[2], v[8]],
                [v[6], v[7], v[8], v[9]],
            ]
        )
        center = np.linalg.solve(-A[:3, :3], v[6:9])
        T = np.eye(4)
        T[3, :3] = center.T

        R = T.dot(A).dot(T.T)
        evals, evecs = np.linalg.eig(R[:3, :3] / -R[3, 3])
        radii = np.sign(evals) * np.sqrt(1.0 / np.abs(evals))
        return cls(radii=radii, center=center, orientations=evecs)

    def sample(
        self, n_samples: int, normal_offset: float = 0.0, **kwargs
    ) -> np.ndarray:
        """
        Samples points from the surface of an ellisoid.

        Parameters
        ----------
        n_samples : int
            Number of samples to draw
        normal_offset : float, optional
            Offset points by normal_offset times their normal vector.

        Returns
        -------
        np.ndarray
            Point coordinates (n, 3).
        """
        radii = self.radii
        if normal_offset is not None:
            radii = [x + normal_offset for x in radii]

        points = Sphere(center=(0, 0, 0), radius=1).sample(n_samples) * radii

        # For each point, find lambda such that the point lies on ellipsoid
        # This is solving: (λx)²/a² + (λy)²/b² + (λz)²/c² = 1
        lambda_vals = 1.0 / np.sqrt(np.sum((points / radii) ** 2, axis=1))
        positions_xyz = points * lambda_vals[:, np.newaxis]
        return positions_xyz.dot(self.orientations.T) + self.center

    def compute_normal(self, points: np.ndarray) -> np.ndarray:
        norm_points = (points - self.center).dot(np.linalg.inv(self.orientations.T))

        normals = np.divide(np.multiply(norm_points, 2), np.square(self.radii))
        normals = np.dot(normals, self.orientations.T)
        return _normalize(normals)

    def compute_distance(self, points: np.ndarray, **kwargs) -> float:
        # Approximate as projected deviation from unit sphere
        norm_points = (points - self.center).dot(np.linalg.inv(self.orientations.T))
        norm_points /= np.linalg.norm(norm_points / self.radii, axis=1)[:, None]
        norm_points = np.dot(norm_points, self.orientations.T) + self.center
        return np.linalg.norm(points - norm_points, axis=1)

    def points_per_sampling(
        self, sampling_density: float, normal_offset: float = None
    ) -> int:
        area_points = np.pi * np.square(sampling_density)

        radii = self.radii
        if normal_offset is not None:
            radii = [x + normal_offset for x in radii]

        area_ellipsoid = np.power(radii[0] * radii[1], 1.6075)
        area_ellipsoid += np.power(radii[0] * radii[2], 1.6075)
        area_ellipsoid += np.power(radii[1] * radii[2], 1.6075)

        area_ellipsoid = np.power(np.divide(area_ellipsoid, 3), 1 / 1.6075)
        area_ellipsoid *= 4 * np.pi

        n_points = np.ceil(np.divide(area_ellipsoid, area_points))
        return int(n_points)


class Cylinder(Parametrization):
    """
    Parametrize a point cloud as a cylinder.

    Parameters
    ----------
    centers : np.ndarray
        Center coordinates of the cylinder in X, Y, and Z.
    orientations : np.ndarray
        Orientation matrix (direction vectors).
    radius : float
        Radius of the cylinder.
    height : float
        Height of the cylinder.
    """

    def __init__(
        self,
        centers: np.ndarray,
        orientations: np.ndarray,
        radius: float,
        height: float,
    ):
        self.centers = np.asarray(centers, dtype=np.float64)
        self.orientations = np.asarray(orientations, dtype=np.float64)
        self.radius = float(radius)
        self.height = float(height)

    @staticmethod
    def _compute_initial_guess(
        positions: np.ndarray,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Compute initial guess for cylinder parameters using PCA.

        Parameters
        ----------
        positions : np.ndarray
            Input point cloud positions.

        Returns
        -------
        center : np.ndarray
            Initial guess for cylinder center.
        direction : np.ndarray
            Initial guess for cylinder axis direction.
        radius : float
            Initial guess for cylinder radius.
        """
        center = np.mean(positions, axis=0)
        positions_centered = positions - center

        cov_mat = np.cov(positions_centered, rowvar=False)
        evals, evecs = np.linalg.eigh(cov_mat)

        sort_idx = np.argsort(evals)[::-1]
        evals = evals[sort_idx]
        evecs = evecs[:, sort_idx]

        direction = evecs[:, -1]

        proj_matrix = np.eye(3) - np.outer(direction, direction)
        projected_points = positions_centered @ proj_matrix
        radius = np.mean(np.linalg.norm(projected_points, axis=1))
        return center, direction, radius

    @classmethod
    def fit(cls, positions: np.ndarray, **kwargs) -> "Cylinder":
        """
        Fit a cylinder to point cloud data with improved stability.

        Parameters
        ----------
        positions : np.ndarray
            Input point cloud positions (N x 3).
        **kwargs : dict
            Additional keyword arguments.

        Returns
        -------
        cylinder : Cylinder
            Fitted cylinder instance.
        """
        positions = np.asarray(positions, dtype=np.float64)
        if positions.shape[1] != 3 or len(positions.shape) != 2:
            raise ValueError("Input must be a Nx3 point cloud.")

        center_init, direction_init, radius_init = cls._compute_initial_guess(positions)
        params_init = np.concatenate([center_init, direction_init, [radius_init]])

        def objective(params):
            center = params[:3]
            direction = params[3:6]
            radius = params[6]
            direction = direction / np.linalg.norm(direction)
            diff = positions - center
            proj = np.dot(diff, direction)[:, np.newaxis] * direction
            perp = diff - proj
            distances = np.abs(np.linalg.norm(perp, axis=1) - radius)
            return np.sum(distances**2)

        constraint = {"type": "eq", "fun": lambda params: np.sum(params[3:6] ** 2) - 1}
        result = optimize.minimize(
            objective,
            params_init,
            method="SLSQP",
            constraints=[constraint],
            options={"ftol": 1e-8, "maxiter": 1000},
        )

        if not result.success:
            print("Warning: Optimization did not converge!")

        center = result.x[:3]
        direction = result.x[3:6]
        direction = direction / np.linalg.norm(direction)
        radius = abs(result.x[6])

        projected_heights = np.dot(positions - center, direction)
        height = np.max(projected_heights) - np.min(projected_heights)
        v1 = np.array([1, 0, 0])
        if not np.allclose(direction, [1, 0, 0]):
            v1 = np.array([0, 1, 0])
        v1 = v1 - np.dot(v1, direction) * direction
        v1 = v1 / np.linalg.norm(v1)
        v2 = np.cross(direction, v1)
        orientations = np.column_stack([v1, v2, direction])

        # TODO: Fix the projection offset on result.x[:3]
        center = center_init

        return cls(
            centers=center, orientations=orientations, radius=radius, height=height
        )

    def compute_normal(self, points: np.ndarray) -> np.ndarray:
        """
        Compute surface normals for points on the cylinder.

        Parameters
        ----------
        points : np.ndarray
            Input points to compute normals for.

        Returns
        -------
        normals : np.ndarray
            Computed surface normals.
        """
        diff = np.asarray(points) - self.centers
        axis = self.orientations[:, 2]
        perp = diff - np.dot(diff, axis)[:, np.newaxis] * axis
        norms = np.linalg.norm(perp, axis=1, keepdims=True)
        normals = np.where(norms > 1e-6, perp / norms, axis)
        return _normalize(normals)

    def sample(
        self,
        n_samples: int,
        normal_offset: float = 0.0,
        mesh_init_factor: int = 5,
        **kwargs,
    ) -> np.ndarray:
        """
        Sample points from the surface of a cylinder.

        Parameters
        ----------
        n_samples : int
            Number of samples to draw
        normal_offset : float, optional
            Offset points by normal_offset times their normal vector.
        mesh_init_factor : int, optional
            Number of times the mesh should be initialized for Poisson sampling.
            Five appears to be a reasonable number. Higher values typically yield
            better sampling.

        Returns
        -------
        np.ndarray
            Point coordinates (n, 3).
        """
        radius, height = self.radius, self.height
        if normal_offset is not None:
            radius = radius + normal_offset
            height = height + normal_offset

        base_samples = int(np.ceil(np.sqrt(n_samples)))
        theta = np.linspace(0, 2 * np.pi, base_samples)
        h = np.linspace(-height / 2, height / 2, base_samples)

        mesh = np.asarray(np.meshgrid(theta, h)).reshape(2, -1).T

        # This does not sample the poles so we need _sample_from_chull
        positions_xyz = np.column_stack(
            [
                radius * np.cos(mesh[:, 0]),
                radius * np.sin(mesh[:, 0]),
                mesh[:, 1],
            ]
        )

        positions_xyz = positions_xyz.dot(self.orientations.T) + self.centers
        return _sample_from_chull(
            positions_xyz=positions_xyz,
            mesh_init_factor=mesh_init_factor,
            n_samples=n_samples,
        )

    def points_per_sampling(
        self, sampling_density: float, normal_offset: float = None
    ) -> int:
        area_points = np.square(sampling_density)

        radius, height = self.radius, self.height
        if normal_offset is not None:
            radius, height = radius + normal_offset, height + normal_offset
        area = 2 * radius * (radius + height)
        n_points = np.ceil(np.divide(area, area_points))
        return int(n_points)


class RBF(Parametrization):
    """
    Parametrize a point cloud using radial basis functions.

    Parameters
    ----------
    rbf : scipy.interpolate.Rbf
        Radial basis function interpolator instance.
    direction : str
        Direction of interpolation relative to positions.
    grid: Tuple
        2D interpolation grid ranges.
    """

    def __init__(self, rbf: type, direction: str, grid: Tuple):
        self.rbf = rbf
        self.grid = grid
        self.direction = direction

    @classmethod
    def fit(
        cls,
        positions: np.ndarray,
        direction: str = "xz",
        function="linear",
        smooth=5,
        **kwargs,
    ) -> "RBF":
        """
        Fit a RBF to a set of 3D points.

        Parameters
        ----------
        positions : np.ndarray
            Point coordinates with shape (n x 3)
        direction : str
            Direction of interpolation relative to positions.
        function : str
            Function type to use.
        smooth : int
            Smoothing factor.

        Returns
        -------
        RBF
            Parametrization instance.
        """
        n_positions = positions.shape[0] // 50
        positions = positions[::n_positions]

        swap = (2, 1, 0)
        if direction == "yz":
            swap = (0, 2, 1)
        elif direction == "xy":
            swap = (0, 1, 2)

        sx, sy, sz = swap
        X, Y, Z = positions[:, sx], positions[:, sy], positions[:, sz]
        rbf = interpolate.Rbf(X, Y, Z, function=function, smooth=smooth)

        grid = ((np.min(X), np.max(X)), (np.min(Y), np.max(Y)))
        return cls(rbf=rbf, direction=direction, grid=grid)

    def sample(
        self, n_samples: int, normal_offset: float = 0.0, **kwargs
    ) -> np.ndarray:
        """
        Sample points from the RBF.

        Parameters
        ----------
        n_samples : int
            Number of samples to draw
        normal_offset : float, optional
            Offset points by normal_offset times their normal vector.

        Returns
        -------
        np.ndarray
            Point coordinates (n, 3).
        """
        (xmin, xmax), (ymin, ymax) = self.grid

        n_samples = int(np.ceil(np.sqrt(n_samples)))
        x, y = np.meshgrid(
            np.linspace(xmin, xmax, n_samples), np.linspace(ymin, ymax, n_samples)
        )
        z = self.rbf(x, y)

        positions_xyz = np.vstack((x.ravel(), y.ravel(), z.ravel())).T
        if self.direction == "xz":
            positions_xyz[:, [0, 2]] = positions_xyz[:, [2, 0]]
        elif self.direction == "yz":
            positions_xyz[:, [1, 2]] = positions_xyz[:, [2, 1]]

        if normal_offset != 0:
            positions_xyz = np.add(
                positions_xyz,
                np.multiply(self.compute_normal(positions_xyz), normal_offset),
            )

        return positions_xyz

    def compute_normal(self, points: np.ndarray) -> np.ndarray:
        normals = compute_normals(points, k=15)
        return _normalize(normals)

    def points_per_sampling(self, sampling_density: float, **kwargs) -> int:
        (xmin, xmax), (ymin, ymax) = self.grid
        surface_area = (xmax - xmin) * (ymax - xmin)

        n_points = np.ceil(np.divide(surface_area, sampling_density))
        return int(n_points)


class TriangularMesh(Parametrization):
    """
    Represent a point cloud as triangular mesh.

    Parameters
    ----------
    mesh : open3d.cpu.pybind.geometry.TriangleMesh
        Triangular mesh.
    """

    def __init__(self, mesh, repair: bool = True):
        self.mesh = mesh

        # We make sure the mesh is clean here to avoid segfaults from
        # ill-defined meshes during curvature or distance computation
        if repair:
            self.mesh.remove_non_manifold_edges()
            self.mesh.remove_degenerate_triangles()
            self.mesh.remove_duplicated_triangles()
            self.mesh.remove_unreferenced_vertices()
            self.mesh.remove_duplicated_vertices()

    def to_file(self, file_path):
        o3d.io.write_triangle_mesh(file_path, self.mesh)

    def subset(self, idx):
        all_indices = np.arange(len(self.vertices))
        indices_to_remove = np.setdiff1d(all_indices, idx)

        new_mesh = to_open3d(self.vertices.copy(), self.triangles.copy())
        new_mesh.remove_vertices_by_index(indices_to_remove)
        return TriangularMesh(new_mesh, repair=False)

    @classmethod
    def from_file(cls, file_path):
        return cls(mesh=o3d.io.read_triangle_mesh(file_path))

    def __getstate__(self):
        state = {"vertices": self.vertices, "triangles": self.triangles}

        if self.mesh.has_vertex_normals():
            state["vertex_normals"] = np.asarray(self.mesh.vertex_normals)
        if self.mesh.has_vertex_colors():
            state["vertex_colors"] = np.asarray(self.mesh.vertex_colors)
        if self.mesh.has_triangle_normals():
            state["triangle_normals"] = np.asarray(self.mesh.triangle_normals)
        return state

    def __setstate__(self, state):
        mesh = to_open3d(state["vertices"], state["triangles"])
        attrs = ("vertex_normals", "vertex_colors", "triangle_normals")
        for attr in attrs:
            if attr not in state:
                continue
            setattr(mesh, attr, o3d.utility.Vector3dVector(state.get(attr)))

        self.mesh = mesh

    @property
    def vertices(self):
        return np.asarray(self.mesh.vertices)

    @property
    def triangles(self):
        return np.asarray(self.mesh.triangles)

    @classmethod
    def fit(
        cls,
        positions: np.ndarray,
        radii: Tuple[float] = (5.0,),
        voxel_size: float = 10,
        max_hole_size: float = -1,
        downsample_input: bool = False,
        elastic_weight: float = 1.0,
        curvature_weight: float = 0.0,
        volume_weight: float = 0.0,
        anchoring: float = 1.0,
        boundary_ring: int = 0,
        n_smoothing: int = 5,
        k_neighbors=50,
        **kwargs,
    ):
        radii = np.asarray(radii).reshape(-1)
        radii = radii[radii > 0]

        # Surface reconstruction normal estimation
        positions = np.asarray(positions, dtype=np.float64)

        # Reduce membrane thickness
        voxel_size = np.max(voxel_size)
        if downsample_input:
            positions = com_cluster_points(positions, cutoff=4 * voxel_size)

        pcd = compute_normals(positions, k=k_neighbors, return_pcd=True)
        mesh = o3d.geometry.TriangleMesh.create_from_point_cloud_ball_pivoting(
            pcd, o3d.utility.DoubleVector(radii)
        )

        # Remove noisy small meshes
        clusters, cluster_n, _ = mesh.cluster_connected_triangles()
        clusters = np.asarray(clusters)
        cluster_n = np.asarray(cluster_n)
        cutoff = 0.02 * cluster_n.sum()
        triangles_to_remove = cluster_n[clusters] < cutoff
        mesh.remove_triangles_by_mask(triangles_to_remove)

        # Repair and smooth
        mesh = mesh.remove_non_manifold_edges()
        mesh = mesh.remove_degenerate_triangles()
        mesh = mesh.remove_duplicated_triangles()
        mesh = mesh.remove_unreferenced_vertices()
        mesh = mesh.remove_duplicated_vertices()
        if n_smoothing > 0:
            mesh = mesh.filter_smooth_taubin(number_of_iterations=n_smoothing)

        if np.asarray(mesh.vertices).shape[0] == 0:
            raise ValueError(
                "No vertices for mesh creation. Try increasing ball pivoting radii."
            )

        if max_hole_size == 0:
            return cls(mesh=mesh)

        # Hole triangulation and fairing
        new_vs, new_fs = triangulate_refine_fair(
            vs=np.asarray(mesh.vertices),
            fs=np.asarray(mesh.triangles),
            hole_len_thr=max_hole_size,
            alpha=elastic_weight,
            beta=curvature_weight,
            gamma=volume_weight,
            anchoring=anchoring,
            n_ring=boundary_ring,
        )
        mesh = to_open3d(new_vs, new_fs)
        mesh = mesh.remove_degenerate_triangles()
        if n_smoothing > 0:
            mesh = mesh.filter_smooth_taubin(number_of_iterations=n_smoothing)

        mesh = mesh.compute_vertex_normals()
        return cls(mesh=mesh)

    def sample(
        self,
        n_samples: int,
        mesh_init_factor: bool = None,
        normal_offset: float = 0.0,
        **kwargs,
    ) -> np.ndarray:
        """
        Samples points from the Triangular mesh.

        Parameters
        ----------
        n_samples : int
            Number of samples to draw
        normal_offset : float, optional
            Offset points by normal_offset times their normal vector.
        mesh_init_factor : int, optional
            Number of times the mesh should be initialized for Poisson sampling.
            Five appears to be a reasonable number. Higher values typically yield
            better sampling.

        Returns
        -------
        np.ndarray
            Point coordinates (n, 3).
        """
        mesh = self.mesh
        if normal_offset != 0:
            self.mesh.compute_vertex_normals()
            mesh = to_open3d(
                np.add(self.vertices, normal_offset * np.asarray(mesh.vertex_normals)),
                self.triangles,
            )
        return _sample_from_mesh(mesh, n_samples, mesh_init_factor)

    def _setup_rayscene(self):
        mesh = o3d.t.geometry.TriangleMesh.from_legacy(self.mesh)
        scene = o3d.t.geometry.RaycastingScene()
        scene_id = scene.add_triangles(mesh)
        return scene, scene_id

    def compute_normal(self, points: np.ndarray) -> np.ndarray:
        self.mesh.compute_triangle_normals()

        scene, _ = self._setup_rayscene()
        points_tensor = o3d.core.Tensor(points, dtype=o3d.core.Dtype.Float32)
        closest_info = scene.compute_closest_points(points_tensor)

        return _normalize(closest_info["primitive_normals"].numpy())

    def compute_curvature(
        self, curvature: str = "gaussian", radius: int = 5
    ) -> np.ndarray:
        use_k_ring = True
        if radius < 2:
            radius, use_k_ring = 2, False

        pd1, pd2, pv1, pv2 = igl.principal_curvature(
            self.vertices, self.triangles, radius=radius, use_k_ring=use_k_ring
        )

        curvature = curvature.lower()
        if curvature == "gaussian":
            return pv1 * pv2
        elif curvature == "mean":
            return (pv1 + pv2) / 2
        else:
            raise ValueError("Only 'gaussian' and 'mean' curvature supported.")

    def compute_vertex_normals(self) -> np.ndarray:
        self.mesh.compute_vertex_normals()
        return np.asarray(self.mesh.vertex_normals).copy()

    def points_per_sampling(
        self, sampling_density: float, normal_offset: float = None
    ) -> int:
        area_per_sample = np.pi * np.square(sampling_density)

        mesh = self.mesh
        if normal_offset is not None:
            self.mesh.compute_vertex_normals()
            mesh = to_open3d(
                np.add(self.vertices, normal_offset * np.asarray(mesh.vertex_normals)),
                self.triangles,
            )

        n_points = np.ceil(np.divide(mesh.get_surface_area(), area_per_sample))
        return int(n_points)

    def compute_distance(
        self,
        points: np.ndarray,
        normals: np.ndarray = None,
        return_projection: bool = False,
        return_indices: bool = False,
        return_triangles: bool = False,
        **kwargs,
    ) -> np.ndarray:
        """
        Compute distance to mesh by ray-casting.

        Parameters
        ----------
        points : np.ndarray
            Points to compute distance from or project onto mesh
        normals : np.ndarray, optional
            Normal vectors for projection direction. If None, computes shortest distance.
        return_projection : bool, optional
            Return points projected onto mesh, defaults to False.
        return_indices : bool, optional
            Return vertex indices closest to projection, defaults to False.
        return_triangles : bool, optional
            Return triangles indices hit by raycasting.

        Returns
        -------
        distances : np.ndarray
            Distance to mesh surface for each point.
        projection : np.ndarray, optional
            Projection of each point onto mesh surface.
        indices : np.ndarray, optional
            Closest vertex to projection.
        triangles : np.ndarray, optional
            Triangle indices hit by projection.
        """
        self.mesh.compute_vertex_normals()
        self.mesh.compute_triangle_normals()

        scene, _ = self._setup_rayscene()

        if normals is None:
            ret = scene.compute_closest_points(
                o3d.core.Tensor(points, dtype=o3d.core.Dtype.Float32)
            )
            projected_points = ret["points"].numpy()
            triangle_indices = ret["primitive_ids"].numpy()
        else:
            rays = o3d.core.Tensor(
                np.hstack([points, normals]), dtype=o3d.core.Dtype.Float32
            )
            hits = scene.cast_rays(rays)

            hit_distances = hits["t_hit"].numpy()
            valid_hits = np.logical_and(hit_distances > 0, np.isfinite(hit_distances))

            n_invalid = valid_hits.size - np.sum(valid_hits)
            if n_invalid > 0:
                warnings.warn(
                    f"{n_invalid} of {valid_hits.size} points did not intersect with "
                    "the mesh. Check the accuracy of the associated normal vectors. "
                    "Falling back to Euclidean distance for those cases."
                )

            projected_points = np.copy(points)
            projected_points[valid_hits] += (
                normals[valid_hits] * hit_distances[valid_hits, np.newaxis]
            )
            triangle_indices = hits["primitive_ids"].numpy()

        dist = np.linalg.norm(points - projected_points, axis=1)
        _, vertex_indices = find_closest_points(self.vertices, projected_points, k=1)

        ret = [dist]

        if return_projection:
            ret.append(projected_points)
        if return_indices:
            ret.append(np.array(vertex_indices))
        if return_triangles:
            ret.append(np.array(triangle_indices))

        if len(ret) == 1:
            return ret[0]
        return ret

    def add_projections(
        self,
        projections: np.ndarray,
        triangle_indices: np.ndarray,
        return_indices: bool = False,
    ) -> Tuple["TriangularMesh", np.ndarray]:
        """
        Add projected points to the mesh by splitting triangles.

        Parameters
        ----------
        projections : np.ndarray
            Projections on the mesh surface to add to the mesh.
        triangle_indices : np.ndarray
            Indices of triangles that each point projects onto.
        return_indices : bool, optional
            Whether to return the index of projections in the new mesh.

        Returns
        -------
        mesh : TriangularMesh
            New mesh with valid projections added
        indices : np.ndarray,
            Array of vertex indices for the added points in the new mesh
        """

        keep = np.logical_and(
            triangle_indices >= 0, triangle_indices < self.triangles.shape[0]
        )
        projections = projections[keep]
        triangle_indices = triangle_indices[keep]

        if len(projections) == 0:
            return to_open3d(self.vertices.copy(), self.triangles.copy())

        n_vertices = self.vertices.shape[0]
        vertices = np.vstack((self.vertices, projections))
        new_indices = np.arange(projections.shape[0]) + n_vertices

        triangles = self.triangles
        triangle_to_points = {}
        for i, tri_idx in enumerate(triangle_indices):
            if tri_idx not in triangle_to_points:
                triangle_to_points[tri_idx] = []
            triangle_to_points[tri_idx].append(i)

        new_triangles = []
        processed_triangles = set()
        for tri_idx, point_indices in triangle_to_points.items():
            v1_idx, v2_idx, v3_idx = triangles[tri_idx]

            processed_triangles.add(tri_idx)

            if len(point_indices) == 1:
                new_vertex_idx = new_indices[point_indices[0]]
                new_triangles.append([v1_idx, v2_idx, new_vertex_idx])
                new_triangles.append([v2_idx, v3_idx, new_vertex_idx])
                new_triangles.append([v3_idx, v1_idx, new_vertex_idx])
                continue

            # Complex case: multiple points in the same triangle
            # Create a Delaunay triangulation of the points plus triangle vertices
            tri_vertices = np.array(
                [vertices[v1_idx], vertices[v2_idx], vertices[v3_idx]]
            )

            point_indices = np.array(point_indices)

            tri_points = projections[point_indices]
            tri_point_indices = new_indices[point_indices]
            all_points = np.vstack((tri_vertices, tri_points))

            # Project points into triangle plan for triangulation
            e1 = tri_vertices[1] - tri_vertices[0]
            normal = np.cross(e1, tri_vertices[2] - tri_vertices[0])

            e1 = e1 / np.linalg.norm(e1)
            e2 = np.cross(normal / np.linalg.norm(normal), e1)
            all_points_2d = np.zeros((all_points.shape[0], 2))
            all_points_2d[:, 0] = np.dot(all_points - tri_vertices[0], e1)
            all_points_2d[:, 1] = np.dot(all_points - tri_vertices[0], e2)

            try:
                from scipy.spatial import Delaunay

                tri = Delaunay(all_points_2d)
                all_indices = np.array(
                    [v1_idx, v2_idx, v3_idx] + tri_point_indices.tolist()
                )

                for simplex in tri.simplices:
                    new_triangles.append(all_indices[simplex])

            except Exception as e:
                warnings.warn(f"Encountered {e}. Falling back to star triangulation.")

                # Fallback: Star triangulation adding each point one by one
                current_triangle_indices = [[v1_idx, v2_idx, v3_idx]]
                for i, new_vertex_idx in enumerate(tri_point_indices):
                    next_triangle_indices = []

                    for t in current_triangle_indices:
                        new_triangles.append([t[0], t[1], new_vertex_idx])
                        new_triangles.append([t[1], t[2], new_vertex_idx])
                        new_triangles.append([t[2], t[0], new_vertex_idx])

                    current_triangle_indices = next_triangle_indices

        for i in range(len(triangles)):
            if i not in processed_triangles:
                new_triangles.append(triangles[i].tolist())

        new_mesh = TriangularMesh(to_open3d(vertices, np.array(new_triangles)))
        if return_indices:
            return new_mesh, new_indices
        return new_mesh

    def geodesic_distance(
        self,
        target_vertices: np.ndarray,
        source_vertices: np.ndarray = None,
        k: int = 1,
        return_indices: bool = False,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Compute geodesic distance from target vertices to their k-nearest
        source vertices on the mesh.

        Parameters
        ----------
        target_vertices : np.ndarray
            Target vertex indices for which to compute distances.
        source_vertices : np.ndarray, optional
            Source vertex indices to compute distances from.
            If None, uses all mesh vertices as sources.
        k : int, optional
            Number of closest source vertices to find for each target vertex.
            Minimum and default value is 1 (nearest neighbor).
        return_indices : bool, optional
            Whether to also return an array of closest vertices in terms of geodesic
            distance. Defaults to False.

        Returns
        -------
        distances: ndarray
            Geodesic distances to k closest sources. Shape is (len(target_vertices), k).
        indices: ndarray, optional
            Corresponding source vertex indices. Shape is (len(target_vertices), k).
        """
        if source_vertices is None:
            source_vertices = np.arange(self.vertices.shape[0])

        k = max(1, min(k, len(source_vertices)))
        source_vertices = np.asarray(source_vertices, dtype=np.int32).ravel()
        target_vertices = np.asarray(target_vertices, dtype=np.int32).ravel()

        k_distances, k_indices = [], []
        for i, tgt_idx in enumerate(target_vertices):
            kwargs = {
                "vt": source_vertices[source_vertices != tgt_idx],
                "vs": np.array([tgt_idx], dtype=np.int32),
            }

            # Faster, we only need to find the closest non-trivial neighbor
            if k == 1:
                kwargs = {"vt": kwargs["vs"], "vs": kwargs["vt"]}

            distances = igl.exact_geodesic(v=self.vertices, f=self.triangles, **kwargs)
            distances = np.atleast_1d(distances)

            sorted_indices = np.argsort(distances)[:k]
            k_distances.append(distances[sorted_indices])
            k_indices.append(source_vertices[sorted_indices])

        k_distances = np.asarray(k_distances)
        if return_indices:
            return k_distances, np.asarray(k_indices)
        return k_distances


class PoissonMesh(TriangularMesh):
    @classmethod
    def fit(
        cls,
        positions: np.ndarray,
        voxel_size: float = None,
        depth: int = 9,
        k_neighbors=50,
        smooth_iter=1,
        pointweight=0.1,
        deldist=1.5,
        scale=1.2,
        samplespernode=5.0,
        **kwargs,
    ):
        mesh = poisson_mesh(
            positions=positions,
            voxel_size=voxel_size,
            depth=depth,
            k_neighbors=k_neighbors,
            smooth_iter=smooth_iter,
            pointweight=pointweight,
            deldist=deldist,
            scale=scale,
            samplespernode=samplespernode,
        )
        return cls(mesh=mesh)


class ClusteredBallPivotingMesh(TriangularMesh):
    @classmethod
    def fit(
        cls,
        positions: np.ndarray,
        voxel_size: float = None,
        radius: int = 0,
        k_neighbors=50,
        smooth_iter=1,
        deldist=-1.0,
        creasethr=90,
        **kwargs,
    ):
        from pymeshlab import MeshSet, Mesh, PercentageValue

        voxel_size = 1 if voxel_size is None else voxel_size
        positions = np.divide(np.asarray(positions, dtype=np.float64), voxel_size)

        ms = MeshSet()
        ms.add_mesh(Mesh(positions))
        ms.compute_normal_for_point_clouds(k=k_neighbors, smoothiter=smooth_iter)
        ms.generate_surface_reconstruction_ball_pivoting(
            ballradius=PercentageValue(radius),
            creasethr=creasethr,
        )
        if deldist > 0:
            ms.compute_scalar_by_distance_from_another_mesh_per_vertex(
                measuremesh=1,
                refmesh=0,
                signeddist=False,
            )
            ms.compute_selection_by_condition_per_vertex(condselect=f"(q>{deldist})")
            ms.compute_selection_by_condition_per_face(
                condselect=f"(q0>{deldist} || q1>{deldist} || q2>{deldist})"
            )
            ms.meshing_remove_selected_vertices_and_faces()

        mesh = ms.current_mesh()
        return cls(
            mesh=to_open3d(mesh.vertex_matrix() * voxel_size, mesh.face_matrix())
        )


class ConvexHull(TriangularMesh):
    """
    Represent a point cloud as triangular mesh.

    Parameters
    ----------
    mesh : open3d.cpu.pybind.geometry.TriangleMesh
        Triangular mesh.
    """

    @classmethod
    def fit(
        cls,
        positions: np.ndarray,
        voxel_size: float = None,
        alpha: float = 1,
        elastic_weight: float = 0,
        curvature_weight: float = 0,
        volume_weight: float = 0,
        anchoring: float = 1.0,
        boundary_ring: int = 0,
        resampling_factor: float = 12.0,
        distance_cutoff: float = 2.0,
        **kwargs,
    ):
        voxel_size = 1 if voxel_size is None else voxel_size
        voxel_size = np.max(voxel_size)

        positions = np.asarray(positions, dtype=np.float64)

        scale = positions.max(axis=0)

        pcd = o3d.geometry.PointCloud()
        pcd.points = o3d.utility.Vector3dVector(positions.copy() / scale)
        try:
            with o3d.utility.VerbosityContextManager(o3d.utility.VerbosityLevel.Error):
                mesh = o3d.geometry.TriangleMesh.create_from_point_cloud_alpha_shape(
                    pcd, alpha
                )
        except Exception as e:
            from scipy.spatial import ConvexHull as scConvexHull

            warnings.warn(f"Encountered {e}. Falling back to scConvexHull.")

            hull = scConvexHull(positions, qhull_options="Qs")
            return cls(mesh=to_open3d(positions[hull.vertices], hull.simplices))

        mesh.vertices = o3d.utility.Vector3dVector(
            np.multiply(np.asarray(mesh.vertices), scale)
        )
        mesh = mesh.remove_non_manifold_edges()
        mesh = mesh.remove_degenerate_triangles()
        mesh = mesh.remove_duplicated_triangles()
        mesh = mesh.remove_unreferenced_vertices()
        mesh = mesh.remove_duplicated_vertices()

        # Better compression and guaranteed to be watertight
        if alpha == 1:
            mesh = o3d.t.geometry.TriangleMesh.from_legacy(mesh)
            mesh = mesh.compute_convex_hull()
            mesh = mesh.to_legacy()

        if elastic_weight == curvature_weight == volume_weight == 0:
            return cls(mesh=mesh)

        # Fair vertices that are distant to input points
        mesh = remesh(mesh, resampling_factor * voxel_size)
        vs, fs = np.asarray(mesh.vertices), np.asarray(mesh.triangles)
        distances, _ = find_closest_points(positions, vs)

        vids = np.where(distances > (resampling_factor / distance_cutoff * voxel_size))[
            0
        ]
        if len(vids) == 0:
            return cls(mesh=to_open3d(vs, fs))

        out_vs = fair_mesh(
            vs,
            fs,
            vids=vids,
            alpha=elastic_weight,
            beta=curvature_weight,
            gamma=volume_weight,
            anchoring=anchoring,
            n_ring=boundary_ring,
        )
        return cls(mesh=to_open3d(out_vs, fs))


class FairHull(ConvexHull):
    pass


class FlyingEdges(TriangularMesh):
    """
    Represent a point cloud as triangular mesh.

    Parameters
    ----------
    mesh : open3d.cpu.pybind.geometry.TriangleMesh
        Triangular mesh.
    """

    @classmethod
    def fit(
        cls,
        positions: np.ndarray,
        voxel_size: float = 10,
        **kwargs,
    ):
        import vtk
        from vtk.util import numpy_support

        voxel_size = tuple(voxel_size for _ in range(positions.shape[1]))

        _volume, offset = points_to_volume(positions, voxel_size, use_offset=True)

        pad = tuple(1 if x == 1 else 0 for x in _volume.shape)
        if any(pad):
            full_pad = tuple((0, x) for x in pad)
            _volume = np.pad(_volume, full_pad)

        volume = vtk.vtkImageData()
        volume.SetSpacing(voxel_size)
        volume.SetDimensions(_volume.shape)
        volume.AllocateScalars(vtk.VTK_FLOAT, 1)
        _volume = numpy_support.numpy_to_vtk(
            _volume.ravel(order="F"), deep=False, array_type=vtk.VTK_FLOAT
        )
        volume.GetPointData().SetScalars(_volume)

        flying_edges = vtk.vtkFlyingEdges3D()
        flying_edges.SetInputData(volume)
        flying_edges.SetValue(0, 0.5)
        flying_edges.ComputeNormalsOn()
        flying_edges.Update()

        polydata = flying_edges.GetOutput()
        vertices_vtk = polydata.GetPoints().GetData()
        vertices = numpy_support.vtk_to_numpy(vertices_vtk)

        polys = polydata.GetPolys()
        cells = numpy_support.vtk_to_numpy(polys.GetData())

        vertices = np.add(vertices, offset * voxel_size)

        faces = cells.reshape(-1, 4)[:, 1:]
        return cls(mesh=to_open3d(vertices, faces), repair=False)


class SplineCurve(Parametrization):
    """
    Parametrize a point cloud as a spline curve.

    Parameters
    ----------
    positions : np.ndarray
        Control points defining the spline curve
    """

    def __init__(self, positions: np.ndarray, order: int = 1, **kwargs):
        self.positions = np.asarray(positions)

        params = self._compute_params()
        if order == 3:
            self._splines = [
                interpolate.CubicSpline(params, self.positions[:, i])
                for i in range(self.positions.shape[1])
            ]
        else:
            self._splines = [
                interpolate.UnivariateSpline(params, self.positions[:, i], k=order)
                for i in range(self.positions.shape[1])
            ]

    def _compute_params(self) -> np.ndarray:
        diff = np.diff(self.positions, axis=0)
        chord_lengths = np.linalg.norm(diff, axis=1)
        cumulative = np.concatenate(([0], np.cumsum(chord_lengths)))
        return cumulative / cumulative[-1]

    @classmethod
    def fit(cls, positions: np.ndarray, **kwargs) -> "SplineCurve":
        return cls(positions=np.asarray(positions, dtype=np.float64), **kwargs)

    def sample(
        self, n_samples: int, normal_offset: float = 0.0, **kwargs
    ) -> np.ndarray:
        t = np.linspace(0, 1, n_samples)
        positions_xyz = np.column_stack([spline(t) for spline in self._splines])

        if normal_offset != 0:
            normals = self.compute_normal(positions_xyz)
            positions_xyz = np.add(positions_xyz, np.multiply(normals, normal_offset))

        return positions_xyz

    def compute_normal(self, points: np.ndarray) -> np.ndarray:
        params = np.linspace(0, 1, len(points))
        tangents = np.column_stack(
            [spline.derivative()(params) for spline in self._splines]
        )
        tangents /= np.linalg.norm(tangents, axis=1)[:, np.newaxis]
        normals = np.zeros_like(tangents)
        normals[:, 0] = -tangents[:, 1]
        normals[:, 1] = tangents[:, 0]
        return _normalize(normals)

    def points_per_sampling(self, sampling_density: float, **kwargs) -> int:
        curve_points = self.sample(1000)
        segments = curve_points[1:] - curve_points[:-1]
        length = np.sum(np.linalg.norm(segments, axis=1))
        n_points = int(np.ceil(length / sampling_density))
        return n_points


PARAMETRIZATION_TYPE = {
    "sphere": Sphere,
    "ellipsoid": Ellipsoid,
    "cylinder": Cylinder,
    "mesh": TriangularMesh,
    "clusterballpivoting": ClusteredBallPivotingMesh,
    "poissonmesh": PoissonMesh,
    "rbf": RBF,
    "convexhull": ConvexHull,
    "spline": SplineCurve,
    "flyingedges": FlyingEdges,
}



from qtpy.QtGui import QAction
from qtpy.QtCore import Qt, QSize, Signal
from qtpy.QtWidgets import (
    QToolBar,
    QWidget,
    QVBoxLayout,
    QLabel,
    QHBoxLayout,
    QToolButton,
    QMenu,
    QPushButton,
    QFrame,
    QFormLayout,
    QSizePolicy,
)
import qtawesome as qta

from .settings import create_setting_widget, get_layout_widget_value
from ..stylesheets import QPushButton_style, QToolButton_style


class SettingsToolButton(QToolButton):
    def __init__(
        self, text, icon_name, settings_config=None, parent=None, callback=None
    ):
        super().__init__(parent)
        self.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextUnderIcon)
        self.setPopupMode(QToolButton.ToolButtonPopupMode.MenuButtonPopup)
        self.setIcon(qta.icon(icon_name, color="#696c6f"))
        self.setText(text)

        self.callback = callback
        self.main_action = QAction(self.icon(), text, self)
        if self.callback is not None:
            self.main_action.triggered.connect(self._apply)
        self.setDefaultAction(self.main_action)

        if settings_config is not None:
            self.settings_menu = SettingsMenu(settings_config, parent=self)
            self.settings_menu.settings_applied.connect(self._applied_settings)
            self.setMenu(self.settings_menu)

    def _apply(self):
        settings = self.settings_menu.get_current_settings()
        return self.callback(**settings) if self.callback else None

    def _applied_settings(self, settings):
        return self.callback(**settings) if self.callback else None


class SettingsMenu(QMenu):
    settings_applied = Signal(dict)

    def __init__(self, config, parent=None):
        super().__init__(parent)
        self.config = config.copy()

        self.method_widgets, self.current_method_widgets = {}, []
        if "method_settings" not in self.config:
            self.config["method_settings"] = {}

        self._setup()
        self.setStyleSheet(
            """
            QMenu {
                border: 1px solid #e5e7eb;
                border-radius: 4px;
                padding: 8px;
            }
            QLabel {
                font-weight: 600;
                min-width: 150px;
            }
            QFormLayout {
                spacing: 10px;
            }
        """
            + QPushButton_style
        )
        self.setFocusPolicy(Qt.FocusPolicy.StrongFocus)
        self.setWindowFlags(self.windowFlags() | Qt.WindowType.Popup)

    def get_current_settings(self):
        ret = {}
        if self.method_combo is not None:
            name = self.method_combo.property("parameter")
            ret[name] = self.method_combo.currentText()

        ret.update(get_layout_widget_value(self.general_form))
        if self.method_layout is not None:
            ret.update(get_layout_widget_value(self.method_layout))
        return ret

    def _setup(self):
        self.main_layout = QVBoxLayout(self)
        self.main_layout.setSpacing(8)
        self.main_layout.setContentsMargins(8, 8, 8, 8)

        title = QLabel(
            f"<span style='font-weight: 600;'>{self.config.get('title', '')}</span>"
        )
        self.main_layout.addWidget(title)

        # Setting shared by multiple methods
        self.general_form = QFormLayout()
        self.general_form.setSpacing(8)
        self.general_form.setLabelAlignment(Qt.AlignmentFlag.AlignLeft)
        self.general_form.setFormAlignment(Qt.AlignmentFlag.AlignLeft)

        # Check whether settings differentiate between different methods
        offset, self.method_combo = 0, None
        base_settings = self.config["settings"][0]
        if "options" in base_settings:
            offset = 1
            self.method_combo = create_setting_widget(base_settings)
            self.method_combo.currentTextChanged.connect(self.update_method_settings)
            self.method_combo.setProperty(
                "parameter", base_settings.get("parameter", "method")
            )
            self.general_form.addRow("Method:", self.method_combo)

        for setting in self.config["settings"][offset:]:
            widget = create_setting_widget(setting)
            self.general_form.addRow(f"{setting['label']}:", widget)

        general_container = QWidget()
        general_container.setLayout(self.general_form)
        self.main_layout.addWidget(general_container)

        self.method_layout = None
        if len(self.config.get("method_settings", {})) > 0:
            separator = QFrame()
            separator.setFixedHeight(2)
            separator.setFrameShape(QFrame.Shape.HLine)
            separator.setFrameShadow(QFrame.Shadow.Sunken)
            separator.setStyleSheet("background-color: #e5e7eb;")
            self.main_layout.addWidget(separator)

            self.method_container = QWidget()
            self.method_layout = QFormLayout(self.method_container)
            self.method_layout.setSpacing(8)
            self.method_layout.setLabelAlignment(Qt.AlignmentFlag.AlignLeft)
            self.method_layout.setFormAlignment(Qt.AlignmentFlag.AlignLeft)

            self.main_layout.addWidget(self.method_container)
            self.main_layout.addStretch()

        apply_btn = QPushButton("Apply")
        apply_btn.clicked.connect(self.apply_settings)
        self.main_layout.addWidget(apply_btn)

        if self.method_combo is not None:
            self.update_method_settings(self.method_combo.currentText())

        self.setFocusProxy(apply_btn)
        self.setAttribute(Qt.WidgetAttribute.WA_Hover, True)
        self.setMouseTracking(True)

    def update_method_settings(self, method):
        if self.method_layout is None:
            return None

        while self.method_layout.rowCount() > 0:
            self.method_layout.removeRow(0)

        self.current_method_widgets.clear()
        settings = self.config.get("method_settings", {}).get(method, {})
        for setting in settings:
            widget = create_setting_widget(setting)
            self.method_layout.addRow(f"{setting['label']}:", widget)
            self.current_method_widgets.append(widget)
        self.adjustSize()

    def apply_settings(self):
        settings = self.get_current_settings()
        self.settings_applied.emit(settings)
        self.close()


class RibbonToolBar(QToolBar):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setIconSize(QSize(20, 20))
        self.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextUnderIcon)
        self.setFixedHeight(85)

        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        self.setStyleSheet(
            """
            QToolBar {
                spacing: 16px;
                border-bottom: 1px solid #6b7280;
                padding: 8px;
            }
            QToolButton {
                min-width: 60px;
                padding: 6px px;
                border-radius: 4px;
                font-size: 11px;
            }
            QToolButton:hover {
                background: #1a000000;
            }
        """
        )
        self.setFocusPolicy(Qt.FocusPolicy.NoFocus)

    def add_section(self, title, actions):
        if len(self.actions()) > 0:
            separator = QWidget()
            separator.setFixedWidth(1)
            separator.setStyleSheet(
                """
                background-color: #696c6f;
                margin-top: 4px;
                margin-bottom: 4px;
            """
            )
            self.addWidget(separator)

        section = QWidget()
        section_layout = QVBoxLayout(section)
        section_layout.setContentsMargins(0, 0, 0, 0)
        section_layout.setSpacing(4)

        label = QLabel(title)
        label.setStyleSheet(
            """
            font-size: 11px;
            margin-bottom: 2px;
        """
        )
        section_layout.addWidget(label)

        actions_widget = QWidget()
        actions_layout = QHBoxLayout(actions_widget)
        actions_layout.setContentsMargins(0, 0, 0, 0)
        actions_layout.setSpacing(2)

        for action in actions:
            actions_layout.addWidget(action)

        section_layout.addWidget(actions_widget)
        self.addWidget(section)


def create_button(
    text, icon_name, parent=None, callback=None, tooltip=None, settings_config=None
):
    if settings_config:
        button = SettingsToolButton(
            text, icon_name, settings_config, parent=parent, callback=callback
        )
    else:
        action = QAction(qta.icon(icon_name, color="#696c6f"), text, parent)
        button = QToolButton()
        button.setDefaultAction(action)
        button.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextUnderIcon)
        if callback:
            button.triggered.connect(callback)

    button.setStyleSheet(QToolButton_style)
    button.setIconSize(QSize(20, 20))
    if tooltip:
        button.setToolTip(tooltip)
    return button



from typing import List, Tuple, Literal

import vtk
import numpy as np
from qtpy.QtCore import Qt, QEvent
from qtpy.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QMessageBox,
    QDockWidget,
    QApplication,
    QMainWindow,
)

from ..widgets.ribbon import create_button


class SegmentationTab(QWidget):
    def __init__(self, cdata, ribbon, legend, **kwargs):
        super().__init__()
        self.cdata = cdata
        self.ribbon = ribbon
        self.legend = legend

        self.trimmer = PlaneTrimmer(self.cdata.data)
        self.transfomer = ClusterTransformer(self.cdata.data)
        layout = QVBoxLayout(self)
        layout.setSpacing(5)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.ribbon)

        self.cdata.data.vtk_widget.installEventFilter(self)

    def eventFilter(self, obj, event):
        """Handle Escape key to exit transformer and trimmer mode."""
        if event.type() == QEvent.Type.KeyPress:
            key = event.key()

            if not self.trimmer.active:
                return super().eventFilter(obj, event)

            if key in (Qt.Key.Key_X, Qt.Key.Key_C, Qt.Key.Key_Z):
                axis = {Qt.Key.Key_X: "x", Qt.Key.Key_C: "y", Qt.Key.Key_Z: "z"}[key]

                self.trimmer.align_to_axis(self.trimmer.plane_widget1, f"-{axis}")
                self.trimmer.align_to_axis(self.trimmer.plane_widget2, f"{axis}")
                return True

        return super().eventFilter(obj, event)

    def show_ribbon(self):
        self.ribbon.clear()
        cluster_actions = [
            create_button(
                "Merge",
                "mdi.merge",
                self,
                self.cdata.data.merge,
                "Merge selected clusters",
            ),
            create_button(
                "Remove",
                "fa5s.trash",
                self,
                self.cdata.data.remove,
                "Remove selected clusters",
            ),
            create_button(
                "Select",
                "mdi.chart-histogram",
                self,
                self._show_histogram,
                "Select clusters by size",
            ),
            create_button(
                "Transform",
                "mdi.rotate-right",
                self,
                self._toggle_transform,
                "Transform selected cluster",
            ),
            create_button(
                "Crop",
                "mdi.map-marker-distance",
                self,
                self._distance_crop,
                "Crop by distance",
            ),
        ]
        self.ribbon.add_section("Base Operations", cluster_actions)

        point_actions = [
            create_button(
                "Cluster",
                "mdi.sitemap",
                self,
                self.cdata.data.cluster,
                "Cluster points",
                CLUSTER_SETTINGS,
            ),
            create_button(
                "Outlier",
                "mdi.filter",
                self,
                self.cdata.data.remove_outliers,
                "Remove outliers",
                OUTLIER_SETTINGS,
            ),
            create_button(
                "Normals",
                "mdi.arrow-expand-up",
                self,
                self.cdata.data.compute_normals,
                "Assign normals",
                NORMAL_SETTINGS,
            ),
            create_button(
                "Trim",
                "mdi.scissors-cutting",
                self,
                self._toggle_trimmer,
                "Trim points using planes",
            ),
            create_button(
                "Thin",
                "mdi.dots-horizontal",
                self,
                self.cdata.data.decimate,
                "Reduce cluster to outer, core or inner points.",
                THINNING_SETTINGS,
            ),
            create_button(
                "Downsample",
                "mdi.focus-field-horizontal",
                self,
                self.cdata.data.downsample,
                "Downsample cluster",
                DOWNSAMPLE_SETTINGS,
            ),
        ]
        self.ribbon.add_section("Point Operations", point_actions)

        analysis_actions = [
            create_button(
                "Properties",
                "mdi.poll",
                self,
                self._show_property_dialog,
                "Analyse object properties",
            ),
        ]
        self.ribbon.add_section("Analysis", analysis_actions)

    def _toggle_trimmer(self):
        if self.trimmer.active:
            return self.trimmer.clean()
        return self.trimmer.show()

    def _toggle_transform(self):
        if self.transfomer.active:
            return self.transfomer.clean()
        return self.transfomer.show()

    def _show_histogram(self):
        from ..dialogs import HistogramDialog
        from ..widgets.dock import create_or_toggle_dock

        dialog = HistogramDialog(self.cdata, parent=self)
        create_or_toggle_dock(self, "histogram_dock", dialog)

    def _show_property_dialog(self):
        from ..dialogs import PropertyAnalysisDialog
        from ..widgets.dock import create_or_toggle_dock

        dialog = PropertyAnalysisDialog(self.cdata, self.legend, parent=self)
        create_or_toggle_dock(self, "property_dock", dialog)

    def _distance_crop(self):
        from ..dialogs import DistanceCropDialog
        from ..properties import GeometryProperties

        fits = self.cdata.format_datalist("models")
        clusters = self.cdata.format_datalist("data")

        dialog = DistanceCropDialog(clusters=clusters, fits=fits, parent=self)
        self.cdata.data.render_update.connect(dialog.populate_lists)
        self.cdata.models.render_update.connect(dialog.populate_lists)

        sources, targets, distance, keep_smaller = dialog.get_results()
        if sources is None:
            return -1

        for source in sources:
            dist = GeometryProperties.compute(
                geometry=source,
                property_name="distance",
                queries=targets,
                include_self=True,
            )
            mask = dist >= distance
            if keep_smaller:
                mask = dist < distance

            if mask.sum() == 0:
                QMessageBox.warning(self, "Warning", "No points satisfy cutoff.")

                continue
            self.cdata.data.add(source[mask])

        self.cdata.data.render()


class ClusterTransformer:
    def __init__(self, data):
        self.data = data
        self.geometry = None
        self.transform_widget = None
        self.selected_cluster = None

    @property
    def active(self):
        return self.transform_widget is not None

    def clean(self):
        """Remove the transform widget and clean up resources."""
        if not self.active:
            return None

        self.transform_widget.Off()
        self.transform_widget.SetEnabled(0)

        self.geometry = None
        self.points = None
        self.normals = None

        self.transform_widget = None
        self.selected_cluster = None
        self.data.vtk_widget.GetRenderWindow().Render()

    def show(self):
        geometries = self.data.get_selected_geometries()
        if not geometries:
            return None

        self.setup()

        self.geometry = geometries[0]

        points = self.geometry.points
        mins = np.min(points, axis=0)
        maxs = np.max(points, axis=0)

        bounds = []
        padding = np.maximum(
            np.multiply(maxs - mins, 0.55), 20 * self.geometry.sampling_rate
        )
        for min_val, max_val, pad in zip(mins, maxs, padding):
            bounds.extend([min_val - pad, max_val + pad])

        # Transforms are w.r.t baseline orientation
        self.points = points.copy()
        self.normals = self.geometry.normals.copy()
        self.transform_widget.PlaceWidget(bounds)
        self.transform_widget.On()
        self.data.vtk_widget.GetRenderWindow().Render()

    def setup(self):
        """Create and configure the 3D widget for transformations."""
        if self.active:
            return None

        self.transform_widget = vtk.vtkBoxWidget()
        self.transform_widget.SetInteractor(
            self.data.vtk_widget.GetRenderWindow().GetInteractor()
        )
        self.transform_widget.SetRotationEnabled(True)
        self.transform_widget.SetTranslationEnabled(True)
        self.transform_widget.SetScalingEnabled(False)

        self.transform_widget.AddObserver("InteractionEvent", self.on_transform)

    def on_transform(self, widget, event):
        """Handle transformation updates."""
        if not self.active:
            return None

        t = vtk.vtkTransform()
        widget.GetTransform(t)
        vmatrix = t.GetMatrix()
        matrix = np.eye(4)
        vmatrix.DeepCopy(matrix.ravel(), vmatrix)

        rotation, translation = matrix[:3, :3], matrix[:3, 3]
        only_translate = np.allclose(rotation, np.eye(3), rtol=1e-10)

        new_points = self.points.copy()
        new_normals = self.normals.copy()
        if not only_translate:
            new_points = np.matmul(new_points, rotation.T, out=new_points)
            new_normals = np.matmul(new_normals, rotation.T, out=new_normals)

        new_points = np.add(new_points, translation, out=new_points)
        self.geometry.swap_data(new_points, normals=new_normals)
        self.data.render()


class PlaneTrimmer:
    def __init__(self, data):
        self.data = data
        self.plane1, self.plane2 = None, None

    @property
    def active(self):
        return self.plane1 is not None and self.plane2 is not None

    def clean(self):
        """Remove the widgets."""
        if self.active:
            self.plane_widget1.Off()
            self.plane_widget1.SetEnabled(0)
            self.plane_widget2.Off()
            self.plane_widget2.SetEnabled(0)

        self.plane_widget1, self.plane_widget2 = None, None
        self.plane1, self.plane2 = None, None

    def show(self, state=None):
        if len(self.data.container) == 0:
            return None

        self._setup()
        self.plane_widget1.SetEnabled(self.active)
        self.plane_widget2.SetEnabled(self.active)
        self.data.render_vtk()

    def _setup(self):
        self.plane1 = vtk.vtkPlane()
        self.plane2 = vtk.vtkPlane()
        self.plane_widget1 = self._setup_plane_widget((1, 0.8, 0.8))
        self.plane_widget2 = self._setup_plane_widget((1, 0.8, 0.8))

        self.align_to_axis(self.plane_widget1, "-z")
        self.align_to_axis(self.plane_widget2, "z")

        bounds = self._get_scene_bounds()
        self.plane_widget1.SetOrigin(bounds[0], bounds[2], bounds[4])
        self.plane_widget2.SetOrigin(bounds[0], bounds[2], bounds[5])

    def align_to_axis(self, widget, axis: Literal["x", "y", "z"]):
        """Align plane normal to specified axis."""
        _normal_mapping = {
            "x": (1, 0, 0),
            "y": (0, 1, 0),
            "z": (0, 0, 1),
            "-x": (-1, 0, 0),
            "-y": (0, -1, 0),
            "-z": (0, 0, -1),
        }
        _axis_mapping = {"x": 0, "y": 1, "z": 2}

        axis = axis.lower()
        normal = _normal_mapping.get(axis, None)
        if normal is None:
            return -1

        plane = self.plane1
        bounds = self._get_scene_bounds()
        origin = [bounds[0], bounds[2], bounds[4]]
        if widget == self.plane_widget2:
            plane = self.plane2
            index = _axis_mapping.get(axis, 0)
            origin[index] = bounds[index * 2 + 1]

        plane.SetNormal(normal)
        plane.SetOrigin(origin)
        widget.SetNormal(*normal)
        widget.SetOrigin(origin)
        self._update_selection()

    def _setup_plane_widget(self, color: Tuple[float, float, float]):
        """Setup an interactive widget for the plane."""
        widget = vtk.vtkImplicitPlaneWidget()
        widget.SetInteractor(self.data.vtk_widget.GetRenderWindow().GetInteractor())
        widget.SetPlaceFactor(1.0)

        bounds = self._get_scene_bounds()
        padding = [(b[1] - b[0]) * 0.1 for b in zip(bounds[::2], bounds[1::2])]
        padding = [
            -padding[i // 2] if i % 2 == 0 else padding[i // 2]
            for i in range(len(bounds))
        ]
        widget.PlaceWidget([sum(x) for x in zip(bounds, padding)])

        widget.GetPlaneProperty().SetColor(*color)
        widget.GetPlaneProperty().SetOpacity(0.4)

        widget.GetNormalProperty().SetColor(0.9, 0.9, 0.9)
        widget.GetNormalProperty().SetLineWidth(1)

        widget.GetEdgesProperty().SetColor(0.9, 0.9, 0.9)
        widget.GetEdgesProperty().SetLineWidth(1)

        widget.TubingOff()
        widget.ScaleEnabledOff()
        widget.OutlineTranslationOff()

        def callback(obj, event):
            origin = [0, 0, 0]
            normal = [0, 0, 0]
            obj.GetNormal(normal)
            obj.GetOrigin(origin)

            plane = self.plane2
            if obj == self.plane_widget1:
                plane = self.plane1

            plane.SetNormal(normal)
            plane.SetOrigin(origin)
            self._update_selection()

        widget.AddObserver("InteractionEvent", callback)
        return widget

    def _get_scene_bounds(self) -> List[float]:
        """Get the bounds of all visible geometry in the scene."""
        bounds = [float("inf"), float("-inf")] * 3

        for i in range(len(self.data.container)):
            if not self.data.container.data[i].visible:
                continue

            geom_bounds = self.data.container.data[i]._data.GetBounds()
            for i in range(len(geom_bounds)):
                func = min if i % 2 == 0 else max
                bounds[i] = func(bounds[i], geom_bounds[i])

        if any((abs(x) == float("inf")) for x in bounds):
            print("Could not determine bounding box - using default.")
            bounds = [-50.0, 50.0] * 3

        return bounds

    def _update_selection(self):
        """Update point selection based on current plane positions."""
        self.data.point_selection.clear()

        for i in range(len(self.data.container)):
            geometry = self.data.container.get(i)
            if not geometry.visible:
                continue

            points = geometry.points
            origin1 = np.array(self.plane1.GetOrigin())
            origin2 = np.array(self.plane2.GetOrigin())
            dist1 = np.dot(points - origin1, np.array(self.plane1.GetNormal()))
            dist2 = np.dot(points - origin2, np.array(self.plane2.GetNormal()))
            self.data.point_selection[geometry.uuid] = np.where((dist1 * dist2) < 0)[0]

        self.data.highlight_selected_points(color=None)


THINNING_SETTINGS = {
    "title": "Thinning Settings",
    "settings": [
        {
            "label": "Method",
            "type": "select",
            "options": ["outer", "core", "inner"],
            "default": "core",
            "description": "Retrieve outer/inner hull or core points.",
        },
    ],
    "method_settings": {
        "core": [
            {
                "label": "Radius",
                "parameter": "cutoff",
                "type": "float",
                "description": "Radius of sphere used for thinning.",
                "default": 50,
                "min": 0,
                "max": 1e32,
                "notes": "Larger radius yields coarser structures.",
            },
        ],
    },
}

NORMAL_SETTINGS = {
    "title": "Normal Settings",
    "settings": [
        {
            "label": "Method",
            "type": "select",
            "options": ["Compute", "Flip"],
            "default": "Compute",
            "description": "Compute new or flip direction of existing normals.",
        },
    ],
    "method_settings": {
        "Compute": [
            {
                "label": "Neighbors",
                "parameter": "k",
                "type": "number",
                "description": "Number of neighboring points to consider for normal estimation",
                "min": 3,
                "max": 100,
                "default": 15,
            },
        ],
        "Flip": [],
    },
}

DOWNSAMPLE_SETTINGS = {
    "title": "Downsample Settings",
    "settings": [
        {
            "label": "Method",
            "type": "select",
            "options": ["Radius", "Number"],
            "default": "Radius",
        },
    ],
    "method_settings": {
        "Radius": [
            {
                "label": "Radius",
                "parameter": "voxel_size",
                "type": "float",
                "default": 40.0,
            },
        ],
        "Number": [
            {
                "label": "Number",
                "parameter": "size",
                "type": "number",
                "min": 1,
                "default": 1000,
            },
        ],
    },
}

CLUSTER_SETTINGS = {
    "title": "Cluster Settings",
    "settings": [
        {
            "label": "Method",
            "type": "select",
            "options": [
                "Connected Components",
                "Envelope",
                "Leiden",
                "DBSCAN",
                "K-Means",
                "Birch",
            ],
            "default": "Connected Components",
        },
        {
            "label": "Use Points",
            "parameter": "use_points",
            "type": "boolean",
            "description": "Use spatial coordinates for clustering",
            "default": True,
        },
        {
            "label": "Use Normals",
            "parameter": "use_normals",
            "type": "boolean",
            "description": "Use normal vectors for clustering",
            "default": False,
        },
        {
            "label": "Drop Noise",
            "parameter": "drop_noise",
            "type": "boolean",
            "description": "Drop noise cluster if available.",
            "default": True,
        },
        {
            "label": "Radius",
            "parameter": "downsampling_radius",
            "type": "float",
            "description": "Consider points within radius as one for clustering.",
            "default": -1.0,
            "min": -1.0,
            "max": 1e32,
            "notes": "Defaults to no downsampling.",
        },
    ],
    "method_settings": {
        "Connected Components": [
            {
                "label": "Distance",
                "parameter": "distance",
                "type": "float",
                "description": "Distance between points to be considered connected.",
                "default": -1.0,
                "min": -1.0,
                "max": 1e32,
                "notes": "Defaults to the associated sampling rate of the cluster.",
            },
        ],
        "Envelope": [
            {
                "label": "Distance",
                "parameter": "distance",
                "type": "float",
                "description": "Distance between points to be considered connected.",
                "default": -1.0,
                "min": -1.0,
                "max": 1e32,
                "notes": "Defaults to the associated sampling rate of the cluster.",
            },
        ],
        "Leiden": [
            {
                "label": "Distance",
                "parameter": "distance",
                "type": "float",
                "description": "Distance between points to be considered connected.",
                "default": -1.0,
                "min": -1.0,
                "max": 1e32,
                "notes": "Defaults to the associated sampling rate of the cluster.",
            },
            {
                "label": "Resolution (log10)",
                "parameter": "resolution_parameter",
                "type": "float",
                "description": "Log10 of resolution parameter for graph clustering.",
                "default": -7.3,
                "min": -1e32,
                "max": 1e32,
                "decimals": 8,
                "notes": "Smaller values yield larger clusters. Range: -8 to -2 for membranes.",
            },
        ],
        "DBSCAN": [
            {
                "label": "Distance",
                "parameter": "distance",
                "type": "float",
                "description": "Distance between neighbors in a cluster.",
                "default": 100.0,
            },
            {
                "label": "Min Points",
                "parameter": "min_points",
                "type": "number",
                "description": "Minimum cluster size.",
                "min": 1,
                "default": 500,
            },
        ],
        "K-Means": [
            {
                "label": "Clusters",
                "parameter": "k",
                "type": "number",
                "min": 1,
                "default": 2,
            },
        ],
        "Birch": [
            {
                "label": "Clusters",
                "parameter": "n_clusters",
                "type": "number",
                "description": "Number of clusters to form.",
                "min": 1,
                "default": 3,
            },
            {
                "label": "Threshold",
                "parameter": "threshold",
                "type": "float",
                "description": "Radius for merging subclusters. Lower values create more clusters.",
                "default": 50.0,
            },
            {
                "label": "Branching Factor",
                "parameter": "branching_factor",
                "type": "number",
                "description": "Max subclusters per node. Higher values use more memory.",
                "min": 1,
                "default": 50,
            },
        ],
    },
}

OUTLIER_SETTINGS = {
    "title": "Remove Outlier",
    "settings": [
        {
            "label": "Method",
            "type": "select",
            "options": ["statistical", "eigenvalue"],
            "default": "statistical",
            "description": "Statistical - General outliers. Eigenvalue - Noisy Edges",
        },
        {
            "label": "Neighbors",
            "parameter": "k_neighbors",
            "type": "number",
            "min": 1,
            "default": 10,
            "description": "k-neigbors for estimating local densities.",
        },
        {
            "label": "Threshold",
            "parameter": "thresh",
            "type": "float",
            "default": 0.02,
            "description": "Threshold is sdev for statistical, eigenvalue ratio otherwise.",
        },
    ],
}



"""
Defines KeybindDialog and the underlying KEYBIND_REGISTRY, which
can be modified to change the Keybind Dialog window.

Copyright (c) 2024 European Molecular Biology Laboratory

Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

from qtpy.QtWidgets import (
    QVBoxLayout,
    QDialog,
    QLabel,
    QGridLayout,
    QGroupBox,
)
from ..widgets import DialogFooter
from ..stylesheets import QGroupBox_style, QPushButton_style


KEYBIND_REGISTRY = {
    "Navigation": [
        ("Z", "Set Camera View along Z-axis"),
        ("X", "Set Camera View along X-axis"),
        ("C", "Set Camera View along Y-axis"),
        ("Left Mouse", "Rotate Scene"),
        ("Shift+Left Mouse", "Translate Scene"),
    ],
    "Visualization": [
        ("A", "Toggle Drawing Mode"),
        ("D", "Toggle Renderer Background Color"),
    ],
    "Selection Operations": [
        ("M", "Merge Selected Cluster or Points"),
        ("Delete", "Remove Selected Cluster or Points"),
        ("R", "Toggle Area Selector"),
        ("E", "Toggle Picking Mode"),
        ("S", "Swap Selector to Fits"),
        ("E", "Expand Selection"),
        ("Right Mouse", "Deselect Cluster or Points"),
    ],
    "File Operations": [
        ("Ctrl+N", "New Session"),
        ("Ctrl+O", "Import Files"),
        ("Ctrl+S", "Save Session"),
        ("Ctrl+P", "Save Screenshot"),
        ("Shift+Ctrl+P", "Save Screenshot to Clipboard"),
        ("Ctrl+E", "Export Animation"),
        ("Ctrl+H", "Show Keybinds"),
    ],
}


class KeybindsDialog(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Keybinds")

        self.layout = QVBoxLayout(self)
        self.layout.setContentsMargins(8, 8, 8, 8)

        self.setStyleSheet(QGroupBox_style + QPushButton_style)

    def create_section(self, title, keybinds):
        frame = QGroupBox(title)
        section_layout = QVBoxLayout(frame)

        grid = QGridLayout()
        for row, (key, description) in enumerate(keybinds):
            key_label = QLabel(key)
            desc_label = QLabel(description)
            grid.addWidget(key_label, row, 0)
            grid.addWidget(desc_label, row, 1)

        section_layout.addLayout(grid)
        frame.setLayout(section_layout)
        return frame

    def show(self):
        while self.layout.count():
            child = self.layout.takeAt(0)
            if child.widget():
                child.widget().deleteLater()

        for title, keybinds in KEYBIND_REGISTRY.items():
            self.layout.addWidget(self.create_section(title, keybinds))

        footer = DialogFooter(dialog=self, margin=(0, 8, 0, 0))
        self.layout.addWidget(footer)

        super().show()



"""
Variety of widgets used throughout the GUI.

Copyright (c) 2024 European Molecular Biology Laboratory

Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

import numpy as np
import pyqtgraph as pg
from qtpy.QtCore import Qt, Signal, QLocale
from qtpy.QtGui import QColor, QDoubleValidator
from qtpy.QtWidgets import (
    QWidget,
    QHBoxLayout,
    QVBoxLayout,
    QSlider,
    QLineEdit,
    QLabel,
    QSpinBox,
    QSizePolicy,
    QComboBox,
    QGridLayout,
)

from ..stylesheets import QSlider_style


class RangeSlider(QWidget):
    """A custom slider that allows selecting a range with two handles."""

    rangeChanged = Signal(float, float)

    def __init__(self, orientation=Qt.Orientation.Horizontal, parent=None):
        super().__init__(parent)
        self.orientation = orientation
        self.lower_value = 0
        self.upper_value = 100

        layout = (
            QHBoxLayout() if orientation == Qt.Orientation.Horizontal else QVBoxLayout()
        )
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

        self.lower_slider = QSlider(orientation)
        self.upper_slider = QSlider(orientation)

        # Emulate range slider appearance
        self._apply_slider_styles()

        for slider, value, callback in [
            (self.lower_slider, 0, self._lower_slider_changed),
            (self.upper_slider, 100, self._upper_slider_changed),
        ]:
            slider.setRange(0, 100)
            slider.setValue(value)
            slider.valueChanged.connect(callback)
            layout.addWidget(slider)

    def _apply_slider_styles(self):
        """Apply custom styles to achieve right-side coloring for lower slider."""
        lower_slider_style = (
            QSlider_style
            + """
        QSlider::sub-page:horizontal {
            background: #e2e8f0;
            border-radius: 2px;
        }
        QSlider::sub-page:horizontal:disabled {
            background: #f1f5f9;
        }
        QSlider::add-page:horizontal {
            background: #94a3b8;
            border-radius: 2px;
        }
        QSlider::add-page:horizontal:disabled {
            background: #cbd5e1;
        }
        """
        )

        upper_slider_style = (
            QSlider_style
            + """
        QSlider::sub-page:horizontal {
            background: #94a3b8;
            border-radius: 2px;
        }
        QSlider::sub-page:horizontal:disabled {
            background: #cbd5e1;
        }
        QSlider::add-page:horizontal {
            background: #e2e8f0;
            border-radius: 2px;
        }
        QSlider::add-page:horizontal:disabled {
            background: #f1f5f9;
        }
        """
        )

        self.lower_slider.setStyleSheet(lower_slider_style)
        self.upper_slider.setStyleSheet(upper_slider_style)

    def _lower_slider_changed(self, value):
        if value > self.upper_slider.value():
            self.lower_slider.setValue(self.upper_slider.value())
            return

        self.lower_value = value
        self.rangeChanged.emit(self.lower_value, self.upper_value)

    def _upper_slider_changed(self, value):
        if value < self.lower_slider.value():
            self.upper_slider.setValue(self.lower_slider.value())
            return

        self.upper_value = value
        self.rangeChanged.emit(self.lower_value, self.upper_value)

    def setRange(self, minimum, maximum):
        self.lower_slider.setRange(minimum, maximum)
        self.upper_slider.setRange(minimum, maximum)

    def setValues(self, lower, upper):
        # Block signals to avoid triggering callbacks
        for slider, value in [(self.lower_slider, lower), (self.upper_slider, upper)]:
            slider.blockSignals(True)
            slider.setValue(value)
            slider.blockSignals(False)

        self.lower_value = lower
        self.upper_value = upper


class HistogramWidget(QWidget):
    cutoff_changed = Signal(float, float)

    def __init__(self, parent=None):
        super().__init__(parent)

        self.data = []
        self.min_value = 0
        self.max_value = 1
        self.bin_count = 20

        self.setup_ui()

    def setup_ui(self):
        main_layout = QVBoxLayout(self)

        self.histogram_plot = pg.PlotWidget()
        self.histogram_plot.setBackground(None)
        self.histogram_plot.getAxis("left").setPen(pg.mkPen(color=(0, 0, 0)))
        self.histogram_plot.getAxis("bottom").setPen(pg.mkPen(color=(0, 0, 0)))
        self.histogram_plot.setLabel("left", "Count")
        self.histogram_plot.setLabel("bottom", "Cluster Size")
        self.histogram_plot.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        controls_layout = self._create_controls()
        self.lower_cutoff_line, self.upper_cutoff_line = self._create_cutoff_lines()

        self.range_slider = RangeSlider(Qt.Orientation.Horizontal)
        self.range_slider.rangeChanged.connect(self._on_slider_range_changed)
        self.range_slider.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        self.range_slider.setMinimumHeight(30)
        self.range_slider.setRange(0, 100)

        main_layout.addWidget(self.histogram_plot)
        main_layout.addLayout(controls_layout)
        main_layout.addWidget(self.range_slider)

    def _create_cutoff_lines(self):
        line_configs = [
            {
                "color": QColor(70, 130, 180),
                "callback": lambda: self._handle_cutoff_drag(is_lower=True),
            },
            {
                "color": QColor(220, 70, 70),
                "callback": lambda: self._handle_cutoff_drag(is_lower=False),
            },
        ]

        lines = []
        for config in line_configs:
            line = pg.InfiniteLine(
                angle=90,
                movable=True,
                pen=pg.mkPen(config["color"], width=2, style=Qt.PenStyle.DotLine),
            )
            line.sigDragged.connect(config["callback"])
            self.histogram_plot.addItem(line)
            lines.append(line)

        return lines

    def _create_controls(self):
        """Create all control widgets and layouts"""
        controls_layout = QGridLayout()
        controls_layout.setSpacing(8)
        controls_layout.setContentsMargins(0, 0, 0, 0)

        self.min_value_input = QLineEdit()
        self.max_value_input = QLineEdit()
        self.transform_combo = QComboBox()
        self.bin_count_spinner = QSpinBox()

        widget_width = 80
        for widget in [
            self.min_value_input,
            self.max_value_input,
            self.transform_combo,
            self.bin_count_spinner,
        ]:
            widget.setMinimumWidth(widget_width)

        validator = QDoubleValidator()
        validator.setLocale(QLocale.c())
        self.min_value_input.setValidator(validator)
        self.max_value_input.setValidator(validator)

        self.transform_combo.addItems(["Linear", "Log"])
        self.transform_combo.currentTextChanged.connect(self._draw_histogram)

        self.bin_count_spinner.setRange(5, 100)
        self.bin_count_spinner.setValue(self.bin_count)
        self.bin_count_spinner.valueChanged.connect(self._on_bin_count_changed)

        self.min_value_input.editingFinished.connect(
            lambda: self._handle_input_change(is_lower=True)
        )
        self.max_value_input.editingFinished.connect(
            lambda: self._handle_input_change(is_lower=False)
        )

        controls_layout.addWidget(
            QLabel("Transform:"), 0, 0, Qt.AlignmentFlag.AlignRight
        )
        controls_layout.addWidget(self.transform_combo, 0, 1)
        controls_layout.addWidget(QLabel("Bins:"), 0, 3, Qt.AlignmentFlag.AlignRight)
        controls_layout.addWidget(self.bin_count_spinner, 0, 4)

        controls_layout.addWidget(
            QLabel("Min Value:"), 1, 0, Qt.AlignmentFlag.AlignRight
        )
        controls_layout.addWidget(self.min_value_input, 1, 1)
        controls_layout.addWidget(
            QLabel("Max Value:"), 1, 3, Qt.AlignmentFlag.AlignRight
        )
        controls_layout.addWidget(self.max_value_input, 1, 4)

        return controls_layout

    def update_histogram(self, data):
        """Update the histogram with new data"""
        self.data = np.asarray(data)

        if self.data.size == 0:
            try:
                return self.plot_widget.clear()
            except Exception:
                return None
        return self._draw_histogram()

    def _invert_scaling(self, value):
        if self.transform_combo.currentText().lower() == "log":
            return 10**value
        return value

    def _draw_histogram(self):
        self.histogram_plot.clear()

        data = self.data
        log_scale = self.transform_combo.currentText().lower() == "log"
        if log_scale:
            data = np.log10(self.data[self.data > 0])

        self.min_value = data.min() - 1
        self.max_value = data.max() + 1
        self._update_cutoff_values(self.min_value, None)

        y, x = np.histogram(data, bins=self.bin_count)
        bar_graph = pg.BarGraphItem(
            x=x[:-1],
            height=y,
            width=(x[1] - x[0]) * 0.8,
            brush=QColor(70, 130, 180),
        )
        self.histogram_plot.addItem(bar_graph)

        self.histogram_plot.addItem(self.lower_cutoff_line)
        self.histogram_plot.addItem(self.upper_cutoff_line)

        label = "Cluster Size" + (" (log scale)" if log_scale else "")
        self.histogram_plot.setLabel("bottom", label)

    def _update_cutoff_values(self, lower_value=None, upper_value=None):
        """Central method to update cutoff values and propagate changes to all UI elements."""
        if lower_value is None:
            lower_value = self.lower_cutoff_line.value()
        if upper_value is None:
            upper_value = self.upper_cutoff_line.value()

        range_span = self.max_value - self.min_value
        if range_span <= 0:
            return None

        lower_value = max(lower_value, self.min_value)
        upper_value = min(max(upper_value, lower_value), self.max_value)

        lower_percent = int(((lower_value - self.min_value) / range_span) * 100)
        upper_percent = int(((upper_value - self.min_value) / range_span) * 100)

        block_elements = [
            self.lower_cutoff_line,
            self.upper_cutoff_line,
            self.range_slider,
            self.min_value_input,
            self.max_value_input,
        ]

        for element in block_elements:
            element.blockSignals(True)

        self.lower_cutoff_line.setValue(lower_value)
        self.upper_cutoff_line.setValue(upper_value)

        locale = QLocale.c()
        self.min_value_input.setText(locale.toString(float(lower_value), "d"))
        self.max_value_input.setText(locale.toString(float(upper_value), "d"))
        self.range_slider.setValues(lower_percent, upper_percent)

        for element in block_elements:
            element.blockSignals(False)

        self.cutoff_changed.emit(
            self._invert_scaling(lower_value), self._invert_scaling(upper_value)
        )

    def _on_slider_range_changed(self, lower_percent, upper_percent):
        """Handle range slider value changes."""
        range_span = self.max_value - self.min_value
        lower_value = self.min_value + (lower_percent / 100.0) * range_span
        upper_value = self.min_value + (upper_percent / 100.0) * range_span
        self._update_cutoff_values(lower_value, upper_value)

    def _handle_cutoff_drag(self, is_lower):
        """Handle dragging of cutoff lines."""
        line = self.lower_cutoff_line if is_lower else self.upper_cutoff_line
        if is_lower:
            self._update_cutoff_values(lower_value=line.value())
        else:
            self._update_cutoff_values(upper_value=line.value())

    def _handle_input_change(self, is_lower):
        """Handle changes to either min/max input field."""
        try:
            input_field = self.min_value_input if is_lower else self.max_value_input
            locale = QLocale.c()
            value = locale.toDouble(input_field.text())[0]

            if is_lower:
                return self._update_cutoff_values(lower_value=value)
            return self._update_cutoff_values(upper_value=value)

        except (ValueError, AttributeError):
            line = self.lower_cutoff_line if is_lower else self.upper_cutoff_line
            input_field.setText(str(int(line.value())))

    def _on_bin_count_changed(self, value):
        """Update the number of bins used in the histogram"""
        self.bin_count = value
        self._draw_histogram()



import warnings

from shutil import which
from pathlib import Path
from typing import Tuple
from subprocess import run
from os.path import splitext, join, basename

import numpy as np

# Hopefully stable download links will be available at some point
MODEL_PATHS = {
    "": "",
    # "MemBrain_seg_v9": "/path/to/model_a.ckpt",
    # "MemBrain_seg_v10": "/home/vmaurer/Documents/segmentation/MemBrain_seg_v10_alpha.ckpt",
}

MEMBRAIN_SETTINGS = {
    "title": "Membrain Settings",
    "settings": [
        {
            "type": "PathSelector",
            "label": "Model",
            "parameter": "model_path",
            "choices": MODEL_PATHS,
            "default": None,
            "description": "Path to pre-trained model ckpt file.",
        },
        {
            "type": "number",
            "label": "Window Size",
            "parameter": "window_size",
            "default": 160,
            "min": 32,
            "max": 512,
            "description": "Size used for inference (smaller values use less GPU but give worse results)",
        },
        {
            "type": "text",
            "label": "Input Sampling",
            "parameter": "input_sampling_rate",
            "default": -1.0,
            "min": 0.1,
            "description": "Pixel size of your tomogram.",
            "notes": "Defaults to the pixel size specified in the header.",
        },
        {
            "type": "text",
            "label": "Output Sampling",
            "parameter": "output_sampling_rate",
            "default": -1.0,
            "min": 0.1,
            "description": "Target pixel size for internal rescaling",
            "notes": "Non default values triggers scaling from input to output scaling.",
        },
        {
            "type": "boolean",
            "label": "Clustering",
            "parameter": "clustering",
            "default": True,
            "description": "Compute connected components of the segmentation",
        },
        {
            "type": "boolean",
            "label": "Augmentation",
            "parameter": "test_time_augmentation",
            "default": True,
            "description": "Use 8-fold test time augmentation for better results but slower runtime",
        },
    ],
}


def _stem(path: str) -> str:
    return splitext(basename(path))[0]


def run_membrainseg(
    tomogram_path: str,
    model_path: str,
    out_folder: str = None,
    window_size: int = 160,
    clustering: bool = True,
    input_sampling_rate: Tuple[float] = -1.0,
    output_sampling_rate: Tuple[float] = -1.0,
    test_time_augmentation: bool = True,
):
    from ..formats.parser import load_density

    warnings.warn(
        "Running MemBrain - Corresponding Citation: "
        "[1] Lamm, L. et al. (2024) bioRxiv, doi.org/10.1101/2024.01.05.574336."
    )
    if which("membrain") is None:
        raise ValueError(
            "The 'membrain' executable was not found in PATH. "
            "Please ensure MemBrain is installed and accessible."
        )

    if out_folder is None:
        out_folder = str(Path.home().joinpath("mosaic/segmentations/membrain"))
    Path(out_folder).mkdir(parents=True, exist_ok=True)

    cmd = [
        "membrain",
        "segment",
        "--tomogram-path",
        tomogram_path,
        "--out-folder",
        out_folder,
        "--ckpt-path",
        model_path,
        "--sliding-window-size",
        window_size,
    ]
    if test_time_augmentation:
        cmd.append("--test-time-augmentation")
    else:
        cmd.append("--no-test-time-augmentation")

    if clustering:
        cmd.append("--store-connected-components")
    else:
        cmd.append("--no-store-connected-components")

    input_sampling_rate = np.max(input_sampling_rate)
    output_sampling_rate = np.max(output_sampling_rate)

    if output_sampling_rate > 0:
        cmd.extend(["--out-pixel-size", f"{output_sampling_rate}", "--rescale-patches"])

        if input_sampling_rate < 0:
            input_sampling_rate = np.max(
                load_density(tomogram_path, use_memmap=True).sampling_rate
            )
            print(f"Setting samping rate to {input_sampling_rate}.")

    if input_sampling_rate > 0:
        cmd.extend(["--in-pixel-size", f"{input_sampling_rate}"])

    ret = run([str(x) for x in cmd])
    out_path = join(
        out_folder, f"{_stem(tomogram_path)}_{_stem(model_path)}.ckpt_segmented.mrc"
    )

    if ret.stderr:
        print(ret.stdout)
        print(ret.stderr)
        out_path = None

    return out_path



"""
VTK Scale Bar widget for adding distance indicators to the vtk viewer.

Copyright (c) 2025 European Molecular Biology Laboratory

Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

import vtk


class ScaleBarWidget:
    def __init__(
        self, renderer: vtk.vtkRenderer, interactor: vtk.vtkRenderWindowInteractor
    ):
        self.renderer = renderer
        self.interactor = interactor

        self.scale_actor = vtk.vtkLegendScaleActor()
        self.scale_actor.AllAxesOff()

        # self.scale_actor.BottomAxisVisibilityOn()
        self.scale_actor.LegendVisibilityOn()

        color = (0.41, 0.42, 0.435)
        self.scale_actor.GetLegendLabelProperty().SetColor(*color)
        self.scale_actor.GetLegendLabelProperty().SetFontSize(0)
        self.scale_actor.GetLegendLabelProperty().SetShadow(0)

        self.scale_actor.GetLegendTitleProperty().SetColor(*color)
        self.scale_actor.GetLegendTitleProperty().SetFontSize(14)
        self.scale_actor.GetLegendTitleProperty().SetShadow(0)

        self.hide()

    def show(self):
        self.visible = True
        self.renderer.AddActor(self.scale_actor)
        return self.interactor.GetRenderWindow().Render()

    def hide(self):
        self.visible = False
        try:
            self.renderer.RemoveActor(self.scale_actor)
        except Exception:
            pass
        return self.interactor.GetRenderWindow().Render()

    def set_label_format(self, format_string: str):
        """Set the format string for the scale labels.

        Args:
            format_string: Format string (e.g., "%.1f mm" or "%.2f µm")
        """
        self.scale_actor.SetLabelFormat(format_string)
        self.interactor.GetRenderWindow().Render()

    def set_units(self, units: str):
        """Set the units for the scale bar.

        Args:
            units: Unit string (e.g., "mm", "µm", "nm")
        """
        format_string = f"%.1f {units}"
        self.set_label_format(format_string)



class AnimationExecutor:
    """Executes animations based on their type and parameters"""

    def __init__(self, vtk_widget):
        self.vtk_widget = vtk_widget
        self.animations = []

    def add_animation(self, animation):
        """Add an animation to the executor"""
        self.animations.append(animation)

    def remove_animation(self, animation_id):
        """Remove an animation by ID"""
        self.animations = [a for a in self.animations if id(a) != animation_id]

    def update(self, global_frame):
        """Update all animations for the current global frame"""
        for animation in self.animations:
            animation.update(global_frame)
        self.vtk_widget.GetRenderWindow().Render()



"""
LegendWidget widget for visualization of scalar mappings.

Copyright (c) 2025 European Molecular Biology Laboratory

Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

import vtk


class LegendWidget:
    def __init__(self, renderer, interactor):
        self.renderer = renderer
        self.interactor = interactor

        self.scalar_bar = vtk.vtkScalarBarActor()
        self.scalar_bar.SetBarRatio(0.07)

        self.scalar_bar.SetDrawFrame(False)
        self.scalar_bar.SetLabelFormat("%.3g")
        self.scalar_bar.SetDrawBackground(False)
        self.scalar_bar.UnconstrainedFontSizeOn()
        self.scalar_bar.SetMaximumNumberOfColors(256)

        color = (0.41, 0.42, 0.435)
        label_property = self.scalar_bar.GetLabelTextProperty()
        label_property.SetColor(*color)
        label_property.SetShadow(0)

        title_property = self.scalar_bar.GetTitleTextProperty()
        title_property.SetColor(*color)
        title_property.SetShadow(0)

        self.widget = vtk.vtkScalarBarWidget()
        self.widget.SetInteractor(self.interactor)
        self.widget.SetScalarBarActor(self.scalar_bar)

        self.widget.GetScalarBarRepresentation().SetShowBorder(False)
        self.widget.ProcessEventsOff()

        self.title = None
        self.visible = False
        self.orientation = "vertical"
        self.set_orientation(self.orientation)

        default_lut = vtk.vtkLookupTable()
        default_lut.SetHueRange(0.667, 0.0)
        default_lut.SetSaturationRange(1.0, 1.0)
        default_lut.SetValueRange(1.0, 1.0)
        default_lut.SetNumberOfColors(256)
        default_lut.Build()
        self.set_lookup_table(default_lut)

    def set_lookup_table(self, lut, title=""):
        self.title = title
        self.scalar_bar.SetLookupTable(lut)
        self.scalar_bar.SetTitle(title)

        return self.interactor.Render()

    def set_orientation(self, orientation):
        is_vertical = orientation.lower() == "vertical"

        self.orientation = "vertical"
        if is_vertical:
            self.scalar_bar.SetOrientationToVertical()
            self.scalar_bar.SetTextPositionToPrecedeScalarBar()
            self.scalar_bar.SetTextPad(-4)
        else:
            self.orientation = "horizontal"
            self.scalar_bar.SetOrientationToHorizontal()
            self.scalar_bar.SetTextPositionToSucceedScalarBar()
            self.scalar_bar.SetTextPad(0)

        self.interactor.Render()

    def show(self):
        if self.visible:
            return None

        self.widget.On()
        self.visible = True
        return self.interactor.Render()

    def hide(self):
        if not self.visible:
            return None

        self.widget.Off()
        self.visible = False
        return self.interactor.Render()



"""
Utilities for repair of triangular meshes.

Hole filling and Leipa triangulation were adapted from
https://github.com/kentechx/hole-filling and are distributed under
MIT license. This origin is indicated as reference for the
respective functions.

Copyright (c) 2024 European Molecular Biology Laboratory

Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

from typing import Tuple, Union

import igl
import numpy as np
import scipy.sparse

_epsilon = 1e-16

__all__ = [
    "triangulate_refine_fair",
    "fair_mesh",
    "get_ring_vertices",
    "close_holes",
    "get_mollified_edge_length",
    "harmonic_deformation",
]


def _close_hole(vs: np.ndarray, fs: np.ndarray, hole_vids, fast=True) -> np.ndarray:
    """Close mesh holes with perimeter length below threshold.

    Parameters
    ----------
    vs : ndarray, shape (N, 3)
        Vertex coordinates.
    fs : ndarray, shape (M, 3)
        Face indices.
    hole_vids : ndarray, shape (K, 3)
        Boundary vertex indices.
    fast : bool, optional
        Whether to use fast hole filling. Default is True.

    Returns
    -------
    ndarray, shape (K, 3)
        Face indices of mesh with holes closed.

    References
    ----------
    .. [1] Code adapted from https://github.com/kentechx/hole-filling
    """

    def hash_func(edges):
        # edges: (n, 2)
        edges = np.core.defchararray.chararray.encode(edges.astype("str"))
        edges = np.concatenate(
            [edges[:, 0:1], np.full_like(edges[:, 0:1], "_", dtype=str), edges[:, 1:2]],
            axis=1,
        )
        edges_hash = np.core.defchararray.add(
            np.core.defchararray.add(edges[:, 0], edges[:, 1]), edges[:, 2]
        )
        return edges_hash

    # create edge hash
    if not fast:
        edges = igl.edges(fs)
        edges_hash = hash_func(edges)

    hole_vids = np.array(hole_vids)
    if len(hole_vids) < 3:
        return fs.copy()

    if len(hole_vids) == 3:
        # fill one triangle
        out_fs = np.concatenate([fs, hole_vids[::-1][None]], axis=0)
        return out_fs

    # heuristically divide the hole
    queue = [hole_vids[::-1]]
    out_fs = []
    while len(queue) > 0:
        cur_vids = queue.pop(0)
        if len(cur_vids) == 3:
            out_fs.append(cur_vids)
            continue

        # current hole
        hole_edge_len = np.linalg.norm(vs[np.roll(cur_vids, -1)] - vs[cur_vids], axis=1)
        hole_len = np.sum(hole_edge_len)
        min_concave_degree = np.inf
        tar_i, tar_j = -1, -1
        for i in range(len(cur_vids)):
            eu_dists = np.linalg.norm(vs[cur_vids[i]] - vs[cur_vids], axis=1)
            if not fast:
                # check if the edge exists
                _edges = np.sort(
                    np.stack([np.tile(cur_vids[i], len(cur_vids)), cur_vids], axis=1),
                    axis=1,
                )
                _edges_hash = hash_func(_edges)
                eu_dists[np.isin(_edges_hash, edges_hash, assume_unique=True)] = np.inf

            geo_dists = np.roll(np.roll(hole_edge_len, -i).cumsum(), i)
            geo_dists = np.roll(np.minimum(geo_dists, hole_len - geo_dists), 1)
            concave_degree = eu_dists / (geo_dists**2 + _epsilon)
            concave_degree[i] = -np.inf  # there may exist two duplicate vertices

            _idx = 1
            j = np.argsort(concave_degree)[_idx]
            while (
                min(
                    (j + len(cur_vids) - i) % len(cur_vids),
                    (i + len(cur_vids) - j) % len(cur_vids),
                )
                <= 1
            ):
                _idx += 1
                j = np.argsort(concave_degree)[_idx]

            if concave_degree[j] < min_concave_degree:
                min_concave_degree = concave_degree[j]
                tar_i, tar_j = min(i, j), max(i, j)

        queue.append(cur_vids[tar_i : tar_j + 1])
        queue.append(np.concatenate([cur_vids[tar_j:], cur_vids[: tar_i + 1]]))

    out_fs = np.concatenate([fs, np.array(out_fs)], axis=0)
    return out_fs


def close_holes(
    vs: np.ndarray, fs: np.ndarray, hole_len_thr: float = 10000.0, fast=True
) -> np.ndarray:
    """Close mesh holes with perimeter length below threshold.

    Parameters
    ----------
    vs : ndarray, shape (N, 3)
        Vertex coordinates.
    fs : ndarray, shape (M, 3)
        Face indices.
    hole_len_thr : float, optional
        Maximum perimeter length of holes to close. Default is 10000.0.
    fast : bool, optional
        Whether to use fast hole filling. Default is True.

    Returns
    -------
    ndarray, shape (K, 3)
        Face indices of mesh with holes closed.

    References
    ----------
    .. [1] Code adapted from https://github.com/kentechx/hole-filling
    """
    out_fs = fs.copy()
    while True:
        updated = False
        for b in igl.all_boundary_loop(out_fs):
            hole_edge_len = np.linalg.norm(vs[np.roll(b, -1)] - vs[b], axis=1).sum()
            if len(b) >= 3 and (hole_edge_len <= hole_len_thr or hole_len_thr < 0):
                out_fs = _close_hole(vs, out_fs, b, fast)
                updated = True
        if not updated:
            break

    return out_fs


def get_mollified_edge_length(
    vs: np.ndarray, fs: np.ndarray, mollify_factor=1e-5
) -> np.ndarray:
    """Calculate mollified edge lengths of mesh faces.

    Parameters
    ----------
    vs : ndarray, shape (N, 3)
        Vertex coordinates.
    fs : ndarray, shape (M, 3)
        Face indices.
    mollify_factor : float, optional
        Factor controlling edge length smoothing. Default is 1e-5.

    Returns
    -------
    ndarray, shape (M, 3)
        Mollified edge lengths for each face triangle.

    References
    ----------
    .. [1] Code adapted from https://github.com/kentechx/hole-filling
    """
    lin = igl.edge_lengths(vs, fs)
    if mollify_factor == 0:
        return lin
    delta = mollify_factor * np.mean(lin)
    eps = np.maximum(0, delta - lin[:, 0] - lin[:, 1] + lin[:, 2])
    eps = np.maximum(eps, delta - lin[:, 0] - lin[:, 2] + lin[:, 1])
    eps = np.maximum(eps, delta - lin[:, 1] - lin[:, 2] + lin[:, 0])
    eps = eps.max()
    lin += eps
    return lin


def _create_weights(size, indices, weight):
    ret = np.full(size, fill_value=0.0)
    ret[indices] = weight
    return scipy.sparse.diags(ret)


def get_ring_vertices(V, F, query_vertices, n=1):
    """
    Find n-ring vertices for given vertices in a mesh.
    Args:
        V: #V by 3 vertices array
        F: #F by 3 faces array
        query_vertices: list/array of vertex indices
        n: integer specifying the number of rings (default=1)
    Returns:
        ring_vertices: set of vertex indices up to n rings away
    """
    vertex_set = set(query_vertices)
    if n < 1:
        return vertex_set

    A = igl.adjacency_matrix(F)
    current_ring = set(query_vertices)
    for i in range(n):
        next_ring = set()
        for v in current_ring:
            next_ring.update(A.getrow(v).indices)
        vertex_set.update(next_ring)
        current_ring = next_ring

    return vertex_set


def harmonic_deformation(vs, fs, vids, k=2):
    vs = vs.astype(np.float32)
    fs = fs.astype(np.int32)

    fixed_vertices = np.setdiff1d(np.arange(len(vs)), vids).astype(np.int32)
    target_positions = vs[fixed_vertices]
    out_vs = igl.harmonic(vs, fs, fixed_vertices, target_positions, k)
    return np.ascontiguousarray(out_vs)


def _fair_mesh(
    vs: np.ndarray,
    fs: np.ndarray,
    vids: np.ndarray,
    alpha=0.0,
    anchoring: Union[float, tuple[float, ...]] = 1.0,
    beta=0.0,
    gamma=0.0,
):
    """
    Minimizes vertex displacement and polyharmonic energy of a mesh at vids.

    Parameters
    ----------
    vs : ndarray, shape (N, 3)
        Vertex coordinates.
    fs : ndarray, shape (M, 3)
        Face indices.
    vids: ndarray (k)
        Vertices to optimize
    alpha : float, optional
        k2 polyharmonic (smoothing) weighting factor. Default 0.0.
    anchoring : float or tuple of float
        Position anchoring strength. 0.0 = strong anchoring, 1.0 = no anchoring.
        Can be defined for all axes or per-axis.
    beta : float, optional
        k3 polyharmonic weighting factor. Default 0.0.
    gamma : float, optional
        Internal mesh pressure. Default 0.0.
    """
    L, M = _robust_laplacian(vs, fs)
    Q2 = igl.harmonic_integrated_from_laplacian_and_mass(L, M, 2)
    Q4 = igl.harmonic_integrated_from_laplacian_and_mass(L, M, 3)

    if np.isscalar(anchoring) or len(anchoring) == 1:
        anchoring = [anchoring, anchoring, anchoring]

    anchoring = np.asarray(anchoring)
    if anchoring.size != 3:
        raise ValueError(
            f"Expected anchoring weights to have len 3, got {anchoring.size}"
        )

    if gamma != 0:
        normals = igl.per_vertex_normals(vs, fs)

    # Solve each axis separately with its own anchoring
    axis_range = range(3)
    if np.unique(anchoring).size == 1:
        anchoring = (anchoring[0],)
        axis_range = ((0, 1, 2),)

    out_vs = np.zeros_like(vs)
    for axis, anch in zip(axis_range, anchoring):
        s = _create_weights(len(vs), vids, alpha * anch)
        a = _create_weights(len(vs), vids, anch)
        b = _create_weights(len(vs), vids, beta * anch)

        displacement = M - a * M
        Q = s * Q2 + b * Q4 + displacement
        B = displacement @ vs[:, axis]

        if gamma != 0:
            B += gamma * normals[:, axis]

        out_vs[:, axis] = igl.spsolve(Q, B)
    return out_vs


def fair_mesh(
    vs: np.ndarray,
    fs: np.ndarray,
    vids: np.ndarray,
    alpha=0.0,
    anchoring=1.0,
    beta=0.0,
    gamma=0.0,
    n_ring=0,
):
    """
    Minimizes vertex displacement and polyharmonic energy of a mesh at vids.

    Parameters
    ----------
    vs : ndarray, shape (N, 3)
        Vertex coordinates.
    fs : ndarray, shape (M, 3)
        Face indices.
    vids: ndarray (k)
        Vertices to optimize
    alpha : float, optional
        k2 polyharmonic (smoothing) weighting factor. Default 0.0.
    anchoring : float, optional
        Position anchoring strength. 0.0 = strong anchoring, 1.0 = no anchoring. Default 1.0.
    beta : float, optional
        k3 polyharmonic weighting factor. Default 0.0.
    gamma : float, optional
        Internal mesh pressure. Default 0.0.
    n_ring : int, optional
        n_ring vertices around vids to consider for fairing. Default 0.
    """
    vs_center = np.mean(vs, axis=0)
    vs = vs - vs_center

    vs_scale = np.std(vs)
    vs_scale = np.where(np.abs(vs_scale) <= 1e-6, 1, vs_scale)
    vs = vs / vs_scale

    vids = np.asarray(vids)
    if n_ring > 0:
        vids = np.asarray(list(get_ring_vertices(vs, fs, vids, n=n_ring)))

    kwargs = {
        "fs": fs,
        "vids": vids,
        "alpha": alpha,
        "beta": beta,
        "anchoring": anchoring,
    }
    out_vs = _fair_mesh(vs, **kwargs)
    if gamma != 0:
        # Two step produced more stable results
        out_vs = _fair_mesh(out_vs, gamma=gamma, **kwargs)
    return out_vs * vs_scale + vs_center


def _robust_laplacian(
    vs, fs, mollify_factor=1e-5, weighting_method=None
) -> Tuple[scipy.sparse.csc_matrix, scipy.sparse.csc_matrix]:
    """
    Get a laplacian with intrinsic Delaunay triangulation and intrinsic mollification.

    Parameters
    ----------
    vs : ndarray, shape (N, 3)
        Vertex coordinates.
    fs : ndarray, shape (M, 3)
        Face indices.
    mollify_factor : float, optional
        Factor controlling edge length smoothing. Default is 1e-5.

    Returns
    -------
    ndarray, shape (M, 3)
        Mollified edge lengths for each face triangle.

    References
    ----------
    .. [1] Code copied from https://github.com/kentechx/hole-filling
    .. [2] https://www.cs.cmu.edu/~kmcrane/Projects/NonmanifoldLaplace/NonmanifoldLaplace.pdf
    """
    lin = get_mollified_edge_length(vs, fs, mollify_factor).astype(np.float64)
    lin, fin = igl.intrinsic_delaunay_triangulation(lin, fs)
    L = igl.cotmatrix_intrinsic(lin, fin)
    M = igl.massmatrix_intrinsic(lin, fin, igl.MASSMATRIX_TYPE_VORONOI)

    if weighting_method == "MEAN_CURVATURE":
        M_inv = scipy.sparse.diags(1.0 / M.diagonal())
        Hn = -M_inv @ (L @ vs)

        # Mean curvature magnitude
        W = scipy.sparse.diags(1.0 + np.linalg.norm(Hn, axis=1))
        M = M @ W
    elif weighting_method == "GAUSSIAN_CURVATURE":
        K = igl.gaussian_curvature(vs, fs)
        W = scipy.sparse.diags(1.0 + np.abs(K))
        M = M @ W

    return L, M


def _triangulation_refine_leipa(
    vs: np.ndarray, fs: np.ndarray, fids: np.ndarray, density_factor: float = np.sqrt(2)
):
    """
    Refine triangles using barycentric subdivision and Delaunay triangulation
    using Liepa's hole filling algorithm [1].


    Parameters
    ----------
    vs : ndarray, shape (N, 3)
        Vertex coordinates.
    fs : ndarray, shape (M, 3)
        Face indices.
    fids : ndarray, shape (K,)
        Indices of faces to refine.
    density_factor : float, optional
        Controls subdivision density. Default is sqrt(2).

    Returns
    -------
    out_vs : ndarray, shape (N+P, 3)
        Output vertices, with new vertices appended.
    out_fs : ndarray, shape (M+Q, 3)
        Output faces, with new faces appended.
    FI : ndarray, shape (M,)
        Maps original face indices to refined face indices.
        FI[i] = -1 indicates face i was deleted.

    References
    ----------
    .. [1] Code adapted from https://github.com/kentechx/hole-filling.
    .. [2] Liepa, P. "Filling holes in meshes." (2003)
    """
    out_vs = np.copy(vs)
    out_fs = np.copy(fs)

    if fids is None or len(fids) == 0:
        return out_vs, out_fs, np.arange(len(fs))

    # initialize sigma
    edges = igl.edges(
        np.delete(out_fs, fids, axis=0)
    )  # calculate the edge length without faces to be refined
    edges = np.concatenate([edges, edges[:, [1, 0]]], axis=0)
    edge_lengths = np.linalg.norm(out_vs[edges[:, 0]] - out_vs[edges[:, 1]], axis=-1)
    edge_length_vids = edges[:, 0]
    v_degrees = np.bincount(edge_length_vids, minlength=len(out_vs))
    v_sigma = np.zeros(len(out_vs))
    v_sigma[v_degrees > 0] = (
        np.bincount(edge_length_vids, weights=edge_lengths, minlength=len(out_vs))[
            v_degrees > 0
        ]
        / v_degrees[v_degrees > 0]
    )
    if np.any(v_sigma == 0):
        v_sigma[v_sigma == 0] = np.median(v_sigma[v_sigma != 0])
        # print("Warning: some vertices have no adjacent faces, the refinement may be incorrect.")

    all_sel_fids = np.copy(fids)
    for _ in range(100):
        # calculate sigma of face centers
        vc_sigma = v_sigma[out_fs].mean(axis=1)  # nf

        # check edge length
        s = density_factor * np.linalg.norm(
            out_vs[out_fs[all_sel_fids]].mean(1, keepdims=True)
            - out_vs[out_fs[all_sel_fids]],
            axis=-1,
        )
        cond = np.all(
            np.logical_and(
                s > vc_sigma[all_sel_fids, None], s > v_sigma[out_fs[all_sel_fids]]
            ),
            axis=1,
        )
        sel_fids = all_sel_fids[cond]  # need to subdivide

        if len(sel_fids) == 0:
            break

        # subdivide
        out_vs, added_fs = igl.false_barycentric_subdivision(out_vs, out_fs[sel_fids])

        # update v_sigma after subdivision
        v_sigma = np.concatenate([v_sigma, vc_sigma[sel_fids]], axis=0)
        assert len(v_sigma) == len(out_vs)

        # delete old faces from out_fs and all_sel_fids
        out_fs[sel_fids] = -1
        all_sel_fids = np.setdiff1d(all_sel_fids, sel_fids)

        # add new vertices, faces & update selection
        out_fs = np.concatenate([out_fs, added_fs], axis=0)
        sel_fids = np.arange(len(out_fs) - len(added_fs), len(out_fs))
        all_sel_fids = np.concatenate([all_sel_fids, sel_fids], axis=0)

        # delaunay
        l = get_mollified_edge_length(out_vs, out_fs[all_sel_fids])
        _, add_fs = igl.intrinsic_delaunay_triangulation(
            l.astype("f8"), out_fs[all_sel_fids]
        )
        out_fs[all_sel_fids] = add_fs

    # update FI, remove deleted faces
    FI = np.arange(len(fs))
    FI[out_fs[: len(fs), 0] < 0] = -1
    idx = np.where(FI >= 0)[0]
    FI[idx] = np.arange(len(idx))
    out_fs = out_fs[out_fs[:, 0] >= 0]
    return out_vs, out_fs, FI


def triangulate_refine_fair(
    vs,
    fs,
    hole_len_thr=-1,
    close_hole_fast=True,
    density_factor=np.sqrt(2),
    alpha=0.05,
    beta=0.0,
    gamma=0,
    n_ring: int = 0,
    anchoring: float = 1.0,
):
    """
    Fill and fair holes in triangular meshes.

    Parameters
    ----------
    vs : ndarray, shape (N, 3)
        Vertex coordinates.
    fs : ndarray, shape (M, 3)
        Face indices.
    hole_len_thr : float, optional
        Maximum hole perimeter to fill. Default is -1 (no limit).
    close_hole_fast : bool, optional
        Use fast hole filling. Default is True.
    density_factor : float, optional
        Controls subdivision density. Default is sqrt(2).
    alpha : float, optional
        Weight for membrane energy. Default is 0.05.
    beta : float, optional
        Weight for curvature energy. Default is 0.
    gamma : float, optional
        Volume pressure. Default is 0.
    n_ring : int, optional
        Also refine n_ring vertices for filled in vertices. Default is 0.

    Returns
    -------
    out_vs : ndarray, shape (N+P, 3)
        Output vertices after filling and fairing.
    out_fs : ndarray, shape (M+Q, 3)
        Output faces after filling and fairing.
    """
    vs = np.asarray(vs).copy()
    fs = np.asarray(fs).copy()
    out_fs = close_holes(vs, fs, hole_len_thr, close_hole_fast)
    add_fids = np.arange(len(fs), len(out_fs))

    nv = len(vs)
    vs, fs, FI = _triangulation_refine_leipa(vs, out_fs, add_fids, density_factor)
    vids = np.arange(nv, len(vs))

    # Fair selected parts of the mesh
    vs = fair_mesh(
        vs,
        fs,
        vids,
        alpha=alpha,
        beta=beta,
        gamma=gamma,
        n_ring=n_ring,
        anchoring=anchoring,
    )
    return vs, fs
